# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from operator import attrgetter
from typing import Callable
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

from kats.consts import IRREGULAR_GRANULARITY_ERROR
from kats.consts import TimeSeriesData
from kats.detectors.cusum_model import CUSUMDetectorModel
from kats.detectors.cusum_model import CusumScoreFunction


class TestIncreaseCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        self.scan_window = 24 * 60 * 60  # in seconds
        self.historical_window = 3 * 24 * 60 * 60  # in seconds
        self.test_data_window = 16  # in hours
        self.regression_sum_score = 12  #
        df_increase = pd.DataFrame(
            {
                "ts_value": np.concatenate([np.random.normal(1, 0.2, 156), np.random.normal(1.5, 0.2, 12)]),
                "time": pd.date_range("2020-01-01", periods=168, freq="H"),
            }
        )
        self.tsd = TimeSeriesData(df_increase)
        self.tsd_value_name = self.tsd.value.name
        data = self.tsd[-self.test_data_window :]

        self.model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window)

        self.score_tsd = self.model.fit_predict(data=data, historical_data=self.tsd[: -self.test_data_window]).scores

        self.score_tsd_percentage_change = self.model._predict(
            data=data,
            score_func=CusumScoreFunction.percentage_change.value,
        ).score

        self.score_tsd_z_score = self.model._predict(data=data, score_func=CusumScoreFunction.z_score.value).score

        self.serialized_model = self.model.serialize()

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ("increase_length_match", len, "score_tsd", "test_data_window"),
            (
                # make sure the time series name are the same
                "increase_time_series_name_match",
                lambda x: x.value.name,
                "score_tsd",
                "tsd_value_name",
            ),
            (
                # the regression is detected
                "increase_regression_detected",
                # pyre-ignore[16]: bool has no attribute sum
                lambda x: (x.value > 0).sum(),
                "score_tsd",
                "regression_sum_score",
            ),
            (
                "increase_percentage_change_length_match",
                len,
                "score_tsd_percentage_change",
                "test_data_window",
            ),
            (
                # the regression is detected
                "increase_percentage_change_regression_detected",
                lambda x: (x.value > 0).sum(),
                "score_tsd_percentage_change",
                "regression_sum_score",
            ),
            (
                "increase_z_score_length_match",
                len,
                "score_tsd_z_score",
                "test_data_window",
            ),
            (
                # the regression is detected
                "increase_z_score_regression_detected",
                lambda x: (x.value > 0).sum(),
                "score_tsd_z_score",
                "regression_sum_score",
            ),
        ]
    )
    def test_score_tsd(
        self,
        name: str,
        func_: Callable[[TimeSeriesData], float],
        attr1: str,
        attr2: str,
    ) -> None:
        self.assertEqual(func_(attrgetter(attr1)(self)), attrgetter(attr2)(self))

    def test_serialized_model(self) -> None:
        self.assertIsInstance(self.serialized_model, bytes)

    def test_new_model(self) -> None:
        model_new = CUSUMDetectorModel(self.serialized_model)
        self.assertEqual(model_new, self.model)

    def test_model(self) -> None:
        self.assertNotEqual(
            self.model,
            CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window),
        )


class TestDecreaseCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        scan_window = 24 * 60 * 60  # in seconds
        historical_window = 3 * 24 * 60 * 60  # in seconds
        self.test_data_window = 6  # in hours
        df_decrease = pd.DataFrame(
            {
                "ts_value": np.concatenate([np.random.normal(2, 0.2, 156), np.random.normal(1, 0.2, 12)]),
                "time": pd.date_range("2020-01-01", periods=168, freq="H"),
            }
        )
        tsd = TimeSeriesData(df_decrease)
        data = tsd[-self.test_data_window :]

        model = CUSUMDetectorModel(scan_window=scan_window, historical_window=historical_window)

        _ = model.fit_predict(data=data, historical_data=tsd[: -self.test_data_window]).scores

        self.score_tsd = model._predict(data=data, score_func=CusumScoreFunction.change.value).score

        self.score_tsd_percentage_change = model._predict(
            data=data,
            score_func=CusumScoreFunction.percentage_change.value,
        ).score

        self.score_tsd_z_score = model._predict(data=data, score_func=CusumScoreFunction.z_score.value).score

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ("decrease_length_match", len, "score_tsd"),
            (
                # the regression is detected
                "decrease_regression_detected",
                # pyre-ignore[16]: bool has no attribute sum
                lambda x: (x.value < 0).sum(),
                "score_tsd",
            ),
            (
                "decrease_percentage_change_length_match",
                len,
                "score_tsd_percentage_change",
            ),
            (
                # the regression is detected
                "decrease_percentage_change_regression_detected",
                lambda x: (x.value < 0).sum(),
                "score_tsd_percentage_change",
            ),
            (
                "decrease_z_score_length_match",
                len,
                "score_tsd_z_score",
            ),
            (
                # the regression is detected
                "decrease_z_score_regression_detected",
                lambda x: (x.value < 0).sum(),
                "score_tsd_z_score",
            ),
        ]
    )
    def test_score_tsd(self, name: str, func_: Callable[[TimeSeriesData], float], attr: str) -> None:
        self.assertEqual(func_(attrgetter(attr)(self)), self.test_data_window)


class TestAdhocCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        self.historical_window = 48 * 60 * 60  # in seconds
        self.scan_window = 11 * 60 * 60 + 50  # in seconds
        n = 168
        self.const_0 = 0
        self.const_12 = 12
        self.const_24 = 24
        df_increase = pd.DataFrame(
            {
                "ts_value": np.concatenate(
                    [
                        np.random.normal(1, 0.2, 48),
                        np.random.normal(0.2, 0.1, 12),
                        np.random.normal(1, 0.2, 60),
                        np.random.normal(2, 0.2, 24),
                        np.random.normal(0.9, 0.2, 24),
                    ]
                ),
                "time": pd.date_range("2020-01-01", periods=n, freq="H"),
            }
        )
        self.tsd = TimeSeriesData(df_increase)

        model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window)
        self.score_tsd = model.fit_predict(data=self.tsd).scores

        # test not enough data
        model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window)
        self.score_tsd_not_enough_data = model.fit_predict(data=self.tsd[-4:], historical_data=self.tsd[-8:-4]).scores

        model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=2 * 3600)
        self.score_tsd_fixed_historical_window = model.fit_predict(data=self.tsd[-8:]).scores

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ("adhoc", len, "score_tsd", len, "tsd"),
            (
                # the regression is went away
                "adhoc_regression_went_away",
                lambda x: x.value[-6:].sum(),
                "score_tsd",
                lambda x: x,
                "const_0",
            ),
            (
                # the increase regression is detected
                "adhoc_regression_increased",
                # pyre-ignore[16]: bool has no attribute sum
                lambda x: (x.value > 0.5).sum(),
                "score_tsd",
                lambda x: x,
                "const_24",
            ),
            (
                # the decrease regression is detected
                "adhoc_regression_decreased",
                lambda x: (x.value < -0.45).sum(),
                "score_tsd",
                lambda x: x,
                "const_12",
            ),
            (
                # test not enough data
                "adhoc_not_enough_data_length_match",
                len,
                "score_tsd_not_enough_data",
                lambda x: len(x[-4:]),
                "tsd",
            ),
            (
                # test not enough data
                "adhoc_not_enough_data_zero_sum_score",
                lambda x: x.value.sum(),
                "score_tsd_not_enough_data",
                lambda x: x,
                "const_0",
            ),
            (
                "adhoc_length_match",
                len,
                "score_tsd_fixed_historical_window",
                lambda x: len(x[-8:]),
                "tsd",
            ),
            (
                "adhoc_zero_sum_score",
                lambda x: x.value.sum(),
                "score_tsd_fixed_historical_window",
                lambda x: x,
                "const_0",
            ),
        ]
    )
    def test_score_tsd(
        self,
        name: str,
        func_1: Callable[[TimeSeriesData], float],
        attr1: str,
        func_2: Callable[[TimeSeriesData], float],
        attr2: str,
    ) -> None:
        self.assertEqual(func_1(attrgetter(attr1)(self)), func_2(attrgetter(attr2)(self)))


class TestMissingDataCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        df = pd.DataFrame(
            {
                "ts_value": [0] * 8,
                "time": [
                    "2020-01-01",
                    "2020-01-02",
                    "2020-01-03",
                    "2020-01-04",
                    "2020-01-05",
                    "2020-01-06",
                    "2020-01-08",
                    "2020-01-09",
                ],
            }
        )
        self.tsd = TimeSeriesData(df)
        # We also assume a bad input here
        model = CUSUMDetectorModel(
            scan_window=2 * 24 * 3600,
            historical_window=2 * 24 * 3600,
        )
        self.score_tsd = model.fit_predict(
            data=self.tsd,
        ).scores

    def test_missing_data_length_match(self) -> None:
        self.assertEqual(len(self.score_tsd), len(self.tsd))

    def test_missing_data_value_match(self) -> None:
        self.assertTrue((self.score_tsd.time.values == self.tsd.time.values).all())


class TestStreamingCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        historical_window = 48 * 60 * 60  # in seconds
        scan_window = 12 * 60 * 60  # in seconds
        self.n = 72
        df_increase = pd.DataFrame(
            {
                "ts_value": np.concatenate([np.random.normal(1, 0.2, 60), np.random.normal(1.5, 0.2, 12)]),
                "time": pd.date_range("2020-01-01", periods=self.n, freq="H"),
            }
        )
        tsd = TimeSeriesData(df_increase)
        # Priming the model
        model = CUSUMDetectorModel(historical_window=historical_window, scan_window=scan_window)
        model.fit(data=tsd[:48])
        pre_serialized_model = model.serialize()

        self.anomaly_score = TimeSeriesData(time=pd.Series(), value=pd.Series([], name="ts_value"))
        # feeding 1 new data point a time
        for i in range(48, self.n):
            model = CUSUMDetectorModel(
                serialized_model=pre_serialized_model,
                historical_window=historical_window,
                scan_window=scan_window,
            )
            self.anomaly_score.extend(
                model.fit_predict(data=tsd[i : i + 1], historical_data=tsd[i - 48 : i]).scores,
                validate=False,
            )
            pre_serialized_model = model.serialize()
        self.anomaly_score.validate_data(validate_frequency=True, validate_dimension=False)

    def test_streaming_length_match(self) -> None:
        self.assertEqual(len(self.anomaly_score), self.n - 48)

    def test_streaming_value_match(self) -> None:
        self.assertTrue(8 <= (self.anomaly_score.value > 0).sum() <= 12)


class TestDecomposingSeasonalityCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        historical_window = 10 * 24 * 60 * 60  # in seconds
        scan_window = 12 * 60 * 60  # in seconds
        n = 480
        periodicity = 24
        self.const_0 = 0
        self.const_10 = 10

        df_sin = pd.DataFrame(
            {
                "time": pd.date_range("2020-01-01", periods=n, freq="H"),
                "ts_value": np.concatenate([20 * np.ones(n // 2), 21 * np.ones(n // 2)])
                + 4 * np.sin(2 * np.pi / periodicity * np.arange(0, n)),
            }
        )

        # removing a few data points to test the missing value handling as well
        self.tsd = TimeSeriesData(pd.concat([df_sin[:100], df_sin[103:]]))

        model = CUSUMDetectorModel(
            scan_window=scan_window,
            historical_window=historical_window,
            remove_seasonality=True,
            score_func=CusumScoreFunction.percentage_change,
        )
        self.score_tsd = model.fit_predict(
            data=self.tsd,
        ).scores

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            (
                "decomposing_seasonality_length_match",
                len,
                "score_tsd",
                len,
                "tsd",
                lambda x, y: x == y,
            ),
            (
                # the scores set to zero after about 7 days
                "decomposing_seasonality_score_after_seven_days",
                lambda x: x.value[-72:].sum(),
                "score_tsd",
                lambda x: x,
                "const_0",
                lambda x, y: x == y,
            ),
            (
                # the increase regression is detected and is on for about 7 days
                # statsmodels version difference will result in different STL results
                "decomposing_seasonality_regression_detected",
                # pyre-ignore[16]: bool has no attribute sumiiuivjtlilhgvhfijkngviirdvbggdrh
                lambda x: np.abs((x.value > 0.01).sum() - 168),
                "score_tsd",
                lambda x: x,
                "const_10",
                lambda x, y: x < y,
            ),
            (
                # make sure the time series time are the same
                "decomposing_seasonality_time_series_same",
                lambda x: x.time.values,
                "score_tsd",
                lambda x: x.time.values,
                "tsd",
                lambda x, y: (x == y).all(),
            ),
            (
                # make sure the time series name are the same
                "decomposing_seasonality_time_series_name_check",
                lambda x: x.value.name,
                "score_tsd",
                lambda x: x.value.name,
                "tsd",
                lambda x, y: x == y,
            ),
        ]
    )
    def test_score_tsd(
        self,
        name: str,
        func_1: Callable[[TimeSeriesData], float],
        attr1: str,
        func_2: Callable[[TimeSeriesData], float],
        attr2: str,
        func_sup: Callable[[float, float], bool],
    ) -> None:
        self.assertTrue(func_sup(func_1(attrgetter(attr1)(self)), func_2(attrgetter(attr2)(self))))


class TestMissingDataRemoveSeasonalityCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(0)
        x = np.random.normal(0.5, 3, 998)
        time_val0 = list(pd.date_range(start="2018-02-03 14:59:59", freq="1800s", periods=1000))
        time_val = time_val0[:300] + time_val0[301:605] + time_val0[606:]
        self.tsd = TimeSeriesData(pd.DataFrame({"time": time_val, "value": pd.Series(x)}))

        time_val01 = list(pd.date_range(start="2018-02-03 14:00:04", freq="1800s", periods=1000))
        time_val1 = time_val01[:300] + time_val01[301:605] + time_val01[606:]
        self.tsd1 = TimeSeriesData(pd.DataFrame({"time": time_val1, "value": pd.Series(x)}))

    def test_interpolation(self) -> None:
        # base = 59 * 60 + 59 or = -1
        model = CUSUMDetectorModel(
            scan_window=360000,
            historical_window=360000,
            remove_seasonality=True,
        )
        score_tsd = model.fit_predict(data=self.tsd).scores
        self.assertEqual(len(score_tsd), len(self.tsd))
        self.assertTrue((score_tsd.time.values == self.tsd.time.values).all())

        # base = 4
        model1 = CUSUMDetectorModel(
            scan_window=360000,
            historical_window=360000,
            remove_seasonality=True,
        )
        score_tsd1 = model1.fit_predict(data=self.tsd1).scores
        self.assertEqual(len(score_tsd1), len(self.tsd1))
        self.assertTrue((score_tsd1.time.values == self.tsd1.time.values).all())


class TestRaiseCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        self.historical_window = 48 * 60 * 60  # in seconds
        self.scan_window = 24 * 60 * 60  # in seconds
        df_increase = pd.DataFrame(
            {
                "ts_value": np.concatenate([np.random.normal(1, 0.2, 156), np.random.normal(1.5, 0.2, 12)]),
                "time": pd.date_range("2020-01-01", periods=168, freq="H"),
            }
        )

        self.tsd = TimeSeriesData(df_increase)

    def test_raise_direction(self) -> None:
        with self.assertRaisesRegex(ValueError, "direction can only be right or left"):
            model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window)
            model._time2idx(self.tsd, self.tsd.time.iloc[0], "")

    def test_raise_model_instantiation(self) -> None:
        with self.assertRaisesRegex(
            ValueError,
            "You must provide either serialized model or values for scan_window and historical_window.",
        ):
            _ = CUSUMDetectorModel()

    def test_raise_time_series_freq(self) -> None:
        with self.assertRaisesRegex(ValueError, IRREGULAR_GRANULARITY_ERROR):
            model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window)
            model.fit_predict(
                data=TimeSeriesData(
                    df=pd.DataFrame(
                        {
                            "value": [0] * 8,
                            "time": [
                                "2020-01-01",
                                "2020-01-02",
                                "2020-01-04",
                                "2020-01-05",
                                "2020-01-07",
                                "2020-01-08",
                                "2020-01-10",
                                "2020-01-11",
                            ],
                        }
                    )
                )
            )

    def test_raise_predict_not_implemented(self) -> None:
        with self.assertRaisesRegex(ValueError, r"predict is not implemented, call fit_predict\(\) instead"):
            model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window)
            model.predict(data=self.tsd)


class TestDeltaReturnCUSUMDetectorModel(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        self.scan_window = 24 * 60 * 60  # in seconds
        self.historical_window = 3 * 24 * 60 * 60  # in seconds
        self.test_data_window = 16  # in hours
        self.regression_sum_score = 12  #
        df_increase = pd.DataFrame(
            {
                "ts_value": np.concatenate([np.random.normal(1, 0.2, 156), np.random.normal(1.5, 0.2, 12)]),
                "time": pd.date_range("2020-01-01", periods=168, freq="H"),
            }
        )
        self.tsd = TimeSeriesData(df_increase)
        self.tsd_value_name = self.tsd.value.name
        data = self.tsd[-self.test_data_window :]

        self.model = CUSUMDetectorModel(scan_window=self.scan_window, historical_window=self.historical_window)

        self.delta = self.model.fit_predict(
            data=data, historical_data=self.tsd[: -self.test_data_window]
        ).anomaly_magnitude_ts

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ("change_point_delta", len, "delta"),
        ]
    )
    def test_change_tsd(
        self,
        name: str,
        func_: Callable[[TimeSeriesData], float],
        attr1: str,
        # attr2: str,
    ) -> None:
        self.assertTrue(func_(attrgetter(attr1)(self)) > 0)


class TestCUSUMDetectorModelWindowsErrors(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        control_time = pd.date_range(start="2018-01-06", freq="D", periods=(100))
        control_val = np.random.normal(0, 5, 100)
        self.data_ts = TimeSeriesData(time=control_time, value=pd.Series(control_val))

    def test_errors(self) -> None:
        # case 1: scan_window < 2 * frequency_sec
        model = CUSUMDetectorModel(scan_window=1 * 24 * 60 * 60, historical_window=2 * 24 * 60 * 60)
        with self.assertRaises(ValueError):
            _ = model.fit_predict(data=self.data_ts)

        # case 2: historical_window < 2 * frequency_sec
        model = CUSUMDetectorModel(scan_window=2 * 24 * 60 * 60, historical_window=1 * 24 * 60 * 60)
        with self.assertRaises(ValueError):
            _ = model.fit_predict(data=self.data_ts)

        # case 3: step_window < frequency_sec
        model = CUSUMDetectorModel(
            scan_window=5 * 24 * 60 * 60,
            historical_window=5 * 24 * 60 * 60,
            step_window=24 * 60 * 60 // 2,
        )
        with self.assertRaises(ValueError):
            _ = model.fit_predict(data=self.data_ts)

        # case 4: step_window >= scan_window
        model = CUSUMDetectorModel(
            scan_window=5 * 24 * 60 * 60,
            historical_window=5 * 24 * 60 * 60,
            step_window=20 * 24 * 60 * 60,
        )
        with self.assertRaises(ValueError):
            _ = model.fit_predict(data=self.data_ts)


class TestCUSUMDetectorModelChangeDirection(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        control_time = pd.date_range(start="2018-01-06", freq="D", periods=(100))
        control_val = np.random.normal(0, 5, 100)
        self.data_ts = TimeSeriesData(time=control_time, value=pd.Series(control_val))

    def test_direction_list(self) -> None:
        model = CUSUMDetectorModel(
            scan_window=10 * 24 * 60 * 60,
            historical_window=10 * 24 * 60 * 60,
            threshold=0.01,
            delta_std_ratio=1.0,
            serialized_model=None,
            change_directions=["increase"],
            score_func=CusumScoreFunction.percentage_change,
        )
        anom = model.fit_predict(data=self.data_ts)
        self.assertEqual(len(anom.scores), len(self.data_ts))

    def test_direction_str(self) -> None:
        # case 1: scan_window < 2 * frequency_sec
        model = CUSUMDetectorModel(
            scan_window=10 * 24 * 60 * 60,
            historical_window=10 * 24 * 60 * 60,
            threshold=0.01,
            delta_std_ratio=1.0,
            serialized_model=None,
            change_directions="increase",
            score_func=CusumScoreFunction.percentage_change,
        )
        self.assertEqual(model.change_directions, ["increase"])
        anom = model.fit_predict(data=self.data_ts)
        self.assertEqual(len(anom.scores), len(self.data_ts))


class TestCUSUMDetectorModelIrregularGranularityError(TestCase):
    def setUp(self) -> None:
        np.random.seed(100)
        ts_time = list(pd.date_range(start="2018-01-06 00:00:00", freq="60s", periods=(100)))
        ts_time = ts_time[:60] + ts_time[61:80] + ts_time[81:]
        ts_time[82] = pd.to_datetime("2018-01-06 01:24:02")
        ts_time[85] = pd.to_datetime("2018-01-06 01:27:22")
        ts_val = np.random.normal(0, 5, 98)
        self.data_ts = TimeSeriesData(time=pd.Series(ts_time), value=pd.Series(ts_val))

    def test_irregular_granularity_error(self) -> None:
        model = CUSUMDetectorModel(
            scan_window=10 * 60,
            historical_window=10 * 60,
            threshold=0.01,
            delta_std_ratio=1.0,
            serialized_model=None,
            change_directions=["increase"],
            remove_seasonality=True,
        )

        with self.assertRaisesRegex(
            ValueError,
            IRREGULAR_GRANULARITY_ERROR,
        ):
            _ = model.fit_predict(data=self.data_ts)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""VAR forecasting Model

VAR model is a multivariate extension of the univariate autoregressive (AR) model.
It captures the linear interdependencies between multiple variables using a system of equations.
Each variable depends not only on its own lagged values but also on the lagged values of other variables.
We use the implementation in statsmodels and re-write the API to adapt Kats development style.

    Typical usage example:

    params = VARParams()
    m = VARModel(data=TSData_multi, params=params)
    m.fit()
    res = m.predict(steps=30)
    m.plot()
"""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from statsmodels.tsa.api import VAR
from statsmodels.tsa.vector_ar.var_model import VARResults

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model
from kats.utils.parameter_tuning_utils import get_default_var_parameter_search_space


class VARParams(Params):
    """Parameter class for VAR model

    This is the parameter class for VAR forecasting model which stands
    for Vector Autoregressive Model.

    Attributes:
        maxlags: Maximum number of lags to check for order selection,
                 Defaults to 12 * (nobs/100.)**(1./4)
        method: Estimation method to use
                Defaults to OLS
        ic: Information criterion to use for VAR order selection
            Defaults to None
        trend: “c” - add constant (Default),
            “ct” - constant and trend,
            “ctt” - constant, linear and quadratic trend,
            “n”/“nc” - no constant, no trend
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.maxlags: int = kwargs.get("maxlags", None)
        self.method: str = kwargs.get("method", "ols")
        self.ic: str = kwargs.get("ic", None)
        self.verbose: bool = kwargs.get("verbose", False)
        self.trend: str = kwargs.get("trend", "c")
        logging.debug("Initialized VARParam instance.:{kwargs}".format(kwargs=kwargs))

    def validate_params(self) -> None:
        """Validate the parameters for VAR model"""

        logging.info("Method validate_params() is not implemented.")
        pass


class VARModel(Model[VARParams]):
    """Model class for VAR

    This class provides fit, predict, and plot methods for VAR model

    Attributes:
        data: the input time series data as :class:`kats.consts.TimeSeriesData`
        params: the parameter class defined with `VARParams`
    """

    model: Optional[VARResults] = None
    k_ar: Optional[int] = None
    sigma_u: Optional[np.ndarray] = None
    resid: Optional[np.ndarray] = None
    freq: Optional[str] = None
    alpha: Optional[float] = None
    dates: Optional[pd.DatetimeIndex] = None
    fcst_dict: Optional[Dict[str, Dict[str, Any]]] = None

    def __init__(self, data: TimeSeriesData, params: VARParams) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.DataFrame):
            msg = "Only support multivariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)

    def fit(self, **kwargs: Any) -> None:
        """Fit VAR model"""

        logging.debug("Call fit()")
        if self.params.maxlags is None:
            # pyre-fixme[16]: `Optional` has no attribute `time`.
            self.params.maxlags = int(12 * (len(self.data.time) / 100.0) ** (1.0 / 4))

        # create VAR model
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        var = VAR(self.data.value)
        logging.info("Created VAR model.")

        # fit VAR model
        self.model = model = var.fit(
            maxlags=self.params.maxlags,
            method=self.params.method,
            ic=self.params.ic,
            verbose=self.params.verbose,
            trend=self.params.trend,
        )
        logging.info("Fitted VAR model.")

        self.k_ar = model.k_ar
        self.sigma_u = model.sigma_u
        self.resid = model.resid

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, include_history: bool = False, **kwargs: Any) -> Dict[str, TimeSeriesData]:
        """Predict with the fitted VAR model

        Args:
            steps: Number of time steps to forecast
            include_history: optional, A boolearn to specify whether to include
                historical data. Default is False.
            freq: optional, frequency of timeseries data.
                Defaults to automatically inferring from time index.
            alpha: optional, significance level of confidence interval.
                Defaults to 0.05

        Returns:
            Disctionary of predicted results for each metric. Each metric result
            has following columns: `time`, `fcst`, `fcst_lower`, and `fcst_upper`
        """
        model = self.model
        if model is None:
            raise ValueError("Call fit() before predict().")

        logging.debug(
            "Call predict() with parameters. " "steps:{steps}, kwargs:{kwargs}".format(steps=steps, kwargs=kwargs)
        )
        self.include_history = include_history
        # pyre-fixme[16]: `Optional` has no attribute `time`.
        self.freq = kwargs.get("freq", pd.infer_freq(self.data.time))
        self.alpha = alpha = kwargs.get("alpha", 0.05)

        fcst = model.forecast_interval(y=model.y, steps=steps, alpha=alpha)
        logging.info("Generated forecast data from VAR model.")
        logging.debug("Forecast data: {fcst}".format(fcst=fcst))

        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)
        dates = dates[1:]  # Return correct number of periods
        self.dates = dates

        self.fcst_dict = fcst_dict = {}
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        ts_names = list(self.data.value.columns)

        for i, name in enumerate(ts_names):
            fcst_df = pd.DataFrame(
                {
                    "time": dates,
                    "fcst": fcst[0][:, i],
                    "fcst_lower": fcst[1][:, i],
                    "fcst_upper": fcst[2][:, i],
                },
                copy=False,
            )
            fcst_dict[name] = fcst_df

        if self.include_history:
            try:
                hist_fcst = model.fittedvalues.values
                hist_dates = self.data.time.iloc[-len(hist_fcst) :]
                for i, name in enumerate(ts_names):
                    print(pd.DataFrame({"time": hist_dates, "fcst": hist_fcst[:, i]}, copy=False))
                    fcst_df = pd.concat(
                        [
                            pd.DataFrame(
                                {"time": hist_dates, "fcst": hist_fcst[:, i]},
                                copy=False,
                            ),
                            fcst_dict[name],
                        ],
                        copy=False,
                    )
                    fcst_dict[name] = fcst_df

            except Exception as e:
                msg = "Failed to generate in-sample forecasts for historical data " f"with error message {e}."
                logging.error(msg)
                raise ValueError(msg)

        logging.debug("Return forecast data: {fcst_dict}".format(fcst_dict=self.fcst_dict))
        ret = {k: TimeSeriesData(v) for k, v in fcst_dict.items()}
        return ret

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[int, int]] = None,
        dpi: int = 120,
        forecast_color: str = "#4267B2",
        history_color: str = "k",
        grid: bool = True,
        xlabel: str = "time",
        **kwargs: Any,
    ) -> plt.Axes:
        """Plot forecasted results from VAR model"""
        fcst_dict = self.fcst_dict
        if fcst_dict is None:
            raise ValueError("Call predict() before plot().")
        if ax is not None:
            raise ValueError("VARModel does not support the ax parameter.")
        dates = self.dates
        assert dates is not None
        fcst_dates = dates.to_pydatetime()
        logging.info("Generating chart for forecast result from VAR model.")

        if figsize is None:
            figsize = (10, 6)
        fig, axes = plt.subplots(ncols=2, dpi=dpi, figsize=figsize)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        for ts_name, ax in zip(self.data.value.columns, axes.flat):
            # pyre-fixme[16]: `Optional` has no attribute `time`.
            ax.plot(self.data.time, self.data.value[ts_name], history_color)

            data = fcst_dict[ts_name]
            ax.plot(fcst_dates, data["fcst"], ls="-", c=forecast_color)
            ax.fill_between(
                fcst_dates,
                data["fcst_lower"],
                data["fcst_upper"],
                color=forecast_color,
                alpha=0.2,
            )

            if grid:
                ax.grid(True, which="major", c="gray", ls="-", lw=1, alpha=0.2)
            ax.set_xlabel(xlabel=xlabel)
            ax.set_ylabel(ylabel=ts_name)

        fig.set_tight_layout(True)
        return axes

    def __str__(self) -> str:
        """VAR model as a string

        Returns:
            String representation of the model name
        """

        return "VAR"

    @staticmethod
    # pyre-fixme[15]: `get_parameter_search_space` overrides method defined in
    #  `Model` inconsistently.
    def get_parameter_search_space() -> List[Dict[str, Any]]:
        """Provide a parameter space for VAR model

        Move the implementation of get_parameter_search_space() out of var
        to avoid the massive dependencies of var and huge build size.
        """

        return get_default_var_parameter_search_space()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.models.simple_heuristic_model import SimpleHeuristicModel
from kats.models.simple_heuristic_model import SimpleHeuristicModelParams

test_univariate_df = pd.DataFrame({"ds": pd.date_range("2022-01-01", periods=30), "y": np.arange(1, 31)})
test_univariate_ts = TimeSeriesData(test_univariate_df, time_col_name="ds")


METHODS = ["last", "mean", "median", "percentile"]


class testSimpleHeuristicModel(TestCase):
    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ("last", False),
            ("mean", False),
            ("median", True),
            ("percentile", True),
        ]
    )
    def test_univariate_data(
        self,
        test_name: str,
        include_history: bool,
    ) -> None:
        horizon = 10
        params = SimpleHeuristicModelParams(method=test_name)
        m = SimpleHeuristicModel(data=test_univariate_ts, params=params)
        m.fit()
        fcst_df = m.predict(steps=horizon, include_history=include_history)
        fcst_df = fcst_df.iloc[-horizon:, :]

        if test_name == "last":
            self.assertTrue(test_univariate_df.y[-horizon:].sum() == fcst_df.fcst.sum())
        elif test_name == "mean":
            self.assertTrue(
                np.mean(np.reshape(np.asarray(test_univariate_df.y), [-1, horizon]), 0).sum() == fcst_df.fcst.sum()
            )
        elif test_name == "median":
            self.assertTrue(
                np.median(np.reshape(np.asarray(test_univariate_df.y), [-1, horizon]), 0).sum() == fcst_df.fcst.sum()
            )
        else:
            self.assertTrue(test_name == "percentile")
            # check percentile 95 (default value)
            self.assertTrue(
                np.percentile(np.reshape(np.asarray(test_univariate_df.y), [-1, horizon]), 95, 0).sum()
                == fcst_df.fcst.sum()
            )

            # check percentile 80 (not default)
            params2 = SimpleHeuristicModelParams(method=test_name, quantile=80)
            m2 = SimpleHeuristicModel(data=test_univariate_ts, params=params2)
            m2.fit()
            fcst_df2 = m2.predict(steps=horizon)
            self.assertTrue(
                np.percentile(np.reshape(np.asarray(test_univariate_df.y), [-1, horizon]), 80, 0).sum()
                == fcst_df2.fcst.sum()
            )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""Apply LSTM for time series forecasting.

The LSTM (long short-term memory) model is a recurrent neural network model that
can be used for sequential data.

See https://en.wikipedia.org/wiki/Long_short-term_memory for more info.

We directly adopt the PyTorch implementation and apply the model for time series
forecast. More details for the PyTorch modules are here:
https://pytorch.org/docs/stable/generated/torch.nn.LSTM.html
"""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from sklearn.preprocessing import MinMaxScaler

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model


class LSTMParams(Params):
    """Parameter class for time series LSTM model

    This is the parameter class for time series LSTM model, it currently
    contains three parameters.

    Attributes:
        hidden_size: LSTM hidden unit size
        time_window: Time series sequence length that feeds into the model
        num_epochs: Number of epochs for the training process
    """

    def __init__(self, hidden_size: int, time_window: int, num_epochs: int) -> None:
        super().__init__()
        self.hidden_size = hidden_size
        self.time_window = time_window
        self.num_epochs = num_epochs

        logging.debug(
            "Initialized LSTMParams instance."
            f"hidden_size:{hidden_size}, time_window:{time_window}"
            f" num_epochs:{num_epochs}"
        )

    def validate_params(self) -> None:
        logging.info("Method validate_params() is not implemented.")
        pass


class LSTMForecast(nn.Module):
    """Torch forecast class for time series LSTM model

    This is the forecast class for time series LSTM model inherited from the
    PyTorch module, detailed implementation for the core LSTM and Linear modules
    can be found here:
    https://pytorch.org/docs/stable/generated/torch.nn.LSTM.html
    https://pytorch.org/docs/stable/generated/torch.nn.Linear.html

    Attributes:
        params: A LSTMParams instance for parameters
        input_size: Input unit feature size for the LSTM layer
        output_size: Output unit feature size from the output Linear layer
    """

    hidden_cell: Optional[Tuple[torch.Tensor, torch.Tensor]] = None

    def __init__(self, params: LSTMParams, input_size: int, output_size: int) -> None:
        super().__init__()

        self.lstm = nn.LSTM(input_size=input_size, hidden_size=params.hidden_size)

        self.linear = nn.Linear(in_features=params.hidden_size, out_features=output_size)

    def forward(self, input_seq: torch.Tensor) -> torch.Tensor:
        """The forward method for the LSTM forecast PyTorch module

        Args:
            input_seq: A torch tensor contains the input data sequence for the
            LSTM layer.

        Returns:
            prediction: A torch tensor contains the output prediction from the
            output Linear layer
        """

        lstm_out, self.hidden_cell = self.lstm(input_seq.view(len(input_seq), 1, -1), self.hidden_cell)
        predictions = self.linear(lstm_out.view(len(input_seq), -1))
        return predictions[-1]


class LSTMModel(Model[LSTMParams]):
    """Kats model class for time series LSTM model

    This is the Kats model class for time series forecast using the LSTM model

    Attributes:
        data: :class:`kats.consts.TimeSeriesData`, the input data
        params: A LSTMParams object for the parameters
    """

    train_data_normalized: Optional[torch.FloatTensor] = None
    scaler: Optional[MinMaxScaler] = None
    lr: Optional[float] = None
    model: Optional[LSTMForecast] = None
    freq: Optional[str] = None
    dates: Optional[pd.DatetimeIndex] = None
    y_fcst: Optional[np.ndarray] = None
    y_fcst_lower: Optional[np.ndarray] = None
    y_fcst_upper: Optional[np.ndarray] = None
    fcst_df: Optional[pd.DataFrame] = None

    def __init__(self, data: TimeSeriesData, params: LSTMParams) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = f"Only support univariate time series, but get {type(self.data.value)}."
            logging.error(msg)
            raise ValueError(msg)

    def __setup_data(self) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        """Prepare input data for the LSTM model

        This method will perform a min-max normalization on the input data, then
        output normalized input sequence data and true values for the prediction.
        More details for the normalizer:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html

        Args:
            None

        Returns:
            A list of tuples that include both the input sequence tensor (with
            normalized values) and ground truth value for prediction.
        """

        # pyre-fixme[16]: `Optional` has no attribute `value`.
        train_data = self.data.value.values.astype(float)

        # scaling using MinMaxScaler
        self.scaler = MinMaxScaler(feature_range=(-1, 1))
        train_data_scaled = self.scaler.fit_transform(train_data.reshape(-1, 1))
        # converting to Tensor
        train_data_normalized = torch.FloatTensor(train_data_scaled).view(-1)
        # pyre-fixme[8]: Attribute has type `Optional[FloatTensor]`; used as `Tensor`.
        self.train_data_normalized = train_data_normalized

        # generating sequence
        inout_seq = []

        time_window = self.params.time_window
        for i in range(len(train_data_normalized) - time_window):
            train_seq = train_data_normalized[i : i + time_window]
            train_label = train_data_normalized[i + time_window : i + time_window + 1]
            inout_seq.append((train_seq, train_label))

        return inout_seq

    # pyre-fixme[15]: `fit` overrides method defined in `Model` inconsistently.
    def fit(self, **kwargs: Any) -> "LSTMModel":
        """Fit the LSTM forecast model

        Args:
            None

        Returns:
            The fitted LSTM model object
        """

        logging.debug("Call fit() with parameters." f"kwargs:{kwargs}")

        # learning rate
        self.lr = kwargs.get("lr", 0.001)

        # supports univariate time series, multivariate support in the future
        model = LSTMForecast(params=self.params, input_size=1, output_size=1)
        self.model = model

        # loss function
        loss_function = nn.MSELoss()

        # optimizer
        # pyre-fixme[6]: For 2nd param expected `float` but got `Optional[float]`.
        optimizer = torch.optim.Adam(model.parameters(), lr=self.lr)

        # inout_seq
        train_inout_seq = self.__setup_data()

        for i in range(self.params.num_epochs):
            for seq, labels in train_inout_seq:
                optimizer.zero_grad()
                model.hidden_cell = (
                    torch.zeros(1, 1, self.params.hidden_size),
                    torch.zeros(1, 1, self.params.hidden_size),
                )
                # prediction using input data
                y_pred = model(seq)
                # calculating loss
                single_loss = loss_function(y_pred, labels)
                single_loss.backward()
                optimizer.step()

            if i % 25 == 1:
                logging.info(f"epoch: {i:3} loss: {single_loss.item():10.8f}")

        return self

    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """Prediction function for a multi-step forecast

        Args:
            steps: number of steps for the forecast

        Returns:
            A pd.DataFrame that includes the forecast and confidence interval
        """
        model = self.model
        if model is None:
            raise ValueError("Call fit() before predict()")
        train_data_normalized = self.train_data_normalized
        scaler = self.scaler
        assert train_data_normalized is not None and scaler is not None
        time_window = self.params.time_window
        hidden_size = self.params.hidden_size

        logging.debug("Call predict() with parameters. " f"steps:{steps}, kwargs:{kwargs}")
        # pyre-fixme[16]: `Optional` has no attribute `time`.
        self.freq = kwargs.get("freq", pd.infer_freq(self.data.time))

        model.eval()

        # get last train input sequence
        test_inputs = train_data_normalized[-time_window:].tolist()

        for _ in range(steps):
            seq = torch.FloatTensor(test_inputs[-time_window:])
            with torch.no_grad():
                model.hidden_cell = (
                    torch.zeros(1, 1, hidden_size),
                    torch.zeros(1, 1, hidden_size),
                )
                test_inputs.append(model(seq).item())

        # inverse transform
        fcst_denormalized = scaler.inverse_transform(np.array(test_inputs[time_window:]).reshape(-1, 1)).flatten()
        logging.info("Generated forecast data from LSTM model.")
        logging.debug(f"Forecast data: {fcst_denormalized}")

        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)
        self.dates = dates[dates != last_date]  # Return correct number of periods
        self.y_fcst = fcst_denormalized
        self.y_fcst_lower = fcst_denormalized * 0.95
        self.y_fcst_upper = fcst_denormalized * 1.05

        self.fcst_df = fcst_df = pd.DataFrame(
            {
                "time": self.dates,
                "fcst": self.y_fcst,
                "fcst_lower": self.y_fcst_lower,
                "fcst_upper": self.y_fcst_upper,
            },
            copy=False,
        )

        logging.debug(f"Return forecast data: {self.fcst_df}")

        return fcst_df

    def __str__(self) -> str:
        return "LSTM"

    @staticmethod
    # pyre-fixme[15]: `get_parameter_search_space` overrides method defined in
    #  `Model` inconsistently.
    def get_parameter_search_space() -> List[Dict[str, Any]]:
        """Get default parameter search space for the LSTM model

        Args:
            None

        Returns:
            A dictionary with the default LSTM parameter search space.
        """

        return [
            {
                "name": "hidden_size",
                "type": "choice",
                "values": list(range(1, 500, 10)),
                "value_type": "int",
                "is_ordered": True,
            },
            {
                "name": "time_window",
                "type": "choice",
                "values": list(range(1, 20, 1)),
                "value_type": "int",
                "is_ordered": True,
            },
            {
                "name": "num_epochs",
                "type": "choice",
                "values": list(range(50, 2000, 50)),
                "value_type": "int",
                "is_ordered": True,
            },
        ]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from unittest import TestCase

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.models.holtwinters import HoltWintersModel
from kats.models.holtwinters import HoltWintersParams
from kats.models.sarima import SARIMAModel
from kats.models.sarima import SARIMAParams
from kats.utils.ensemble_predict_interval import ensemble_predict_interval


class testEnsemblePredictInterval(TestCase):
    def setUp(self) -> None:
        # create time series data
        np.random.seed(0)
        val = (
            np.arange(180) / 6
            + np.sin(np.pi * np.arange(180) / 6) * 20
            + +np.cos(np.arange(180)) * 20
            + np.random.randn(180) * 10
        )
        ts = TimeSeriesData(pd.DataFrame({"time": pd.date_range("2021-05-06", periods=180), "val": val}))
        self.test_ts = ts[120:]
        self.hist_ts = ts[:120]

    def test_EPI_Sarima(self) -> None:
        # test EPI on SARIMA model
        params = SARIMAParams(p=2, d=1, q=1, trend="ct", seasonal_order=(1, 0, 1, 12))
        epi = ensemble_predict_interval(
            # pyre-fixme[6]: Incompatible parameter type
            model=SARIMAModel,
            model_params=params,
            ts=self.hist_ts,
            n_block=5,
            ensemble_size=4,
            multiprocessing=True,
        )
        res = epi.get_projection(step=20)
        self.assertEqual(res.shape, (20, 3))
        epi.pi_comparison_plot(self.test_ts)

    def test_EPI_HW(self) -> None:
        # test EPI on Holt-Winters model
        epi = ensemble_predict_interval(
            # pyre-fixme[6]: Incompatible parameter type
            model=HoltWintersModel,
            model_params=HoltWintersParams(),
            ts=self.hist_ts,
            n_block=5,
            ensemble_size=10,
            multiprocessing=True,
        )
        res = epi.get_projection(step=40, rolling_based=True)
        self.assertEqual(res.shape, (40, 3))

        res_other_conf_level = epi.get_fcst_band_with_level(confidence_level=0.5)
        self.assertEqual(res_other_conf_level.shape, (40, 3))

        epi.pi_comparison_plot(self.test_ts)
        # test if there is no test_ts provided
        epi.pi_comparison_plot()

        # fcst step is greater than test set length
        epi2 = ensemble_predict_interval(
            # pyre-fixme[6]: Incompatible parameter type
            model=HoltWintersModel,
            model_params=HoltWintersParams(),
            ts=self.hist_ts,
            n_block=5,
            ensemble_size=4,
        )
        res2 = epi2.get_projection(step=80)
        self.assertEqual(res2.shape, (80, 3))

        res_other_conf_level2 = epi2.get_fcst_band_with_level(confidence_level=0.5)
        self.assertEqual(res_other_conf_level2.shape, (80, 3))

        epi2.pi_comparison_plot(self.test_ts)
        epi2.pi_comparison_plot(self.test_ts, test_data_only=True)

    def test_errors(self) -> None:
        # both block_size and n_block are None
        with self.assertRaises(ValueError):
            _ = ensemble_predict_interval(
                # pyre-fixme[6]: Incompatible parameter type
                model=HoltWintersModel,
                model_params=HoltWintersParams(),
                ts=self.hist_ts,
            )

        # not suitable n_block when block_size is None
        for x in [1, -1, 200]:
            with self.assertRaises(ValueError):
                _ = ensemble_predict_interval(
                    # pyre-fixme[6]: Incompatible parameter type
                    model=HoltWintersModel,
                    model_params=HoltWintersParams(),
                    ts=self.hist_ts,
                    n_block=x,
                )

        # not suitable block_size when n_block is None
        for y in [1, -1, 200]:
            with self.assertRaises(ValueError):
                _ = ensemble_predict_interval(
                    # pyre-fixme[6]: Incompatible parameter type
                    model=HoltWintersModel,
                    model_params=HoltWintersParams(),
                    ts=self.hist_ts,
                    block_size=y,
                )

        # not suitable block_size and n_block vals
        for x, y in [(100, 100), (1, 100), (100, 1)]:
            with self.assertRaises(ValueError):
                _ = ensemble_predict_interval(
                    # pyre-fixme[6]: Incompatible parameter type
                    model=HoltWintersModel,
                    model_params=HoltWintersParams(),
                    ts=self.hist_ts,
                    n_block=x,
                    block_size=y,
                )

        # not suitable ensemble size
        for m in [1, -1, 0]:
            with self.assertRaises(ValueError):
                _ = ensemble_predict_interval(
                    # pyre-fixme[6]: Incompatible parameter type
                    model=HoltWintersModel,
                    model_params=HoltWintersParams(),
                    ts=self.hist_ts,
                    n_block=5,
                    ensemble_size=m,
                )

        # error of plot
        epi = ensemble_predict_interval(
            # pyre-fixme[6]: Incompatible parameter type
            model=HoltWintersModel,
            model_params=HoltWintersParams(),
            ts=self.hist_ts,
            n_block=5,
        )
        with self.assertRaises(ValueError):
            # havn't been trained
            epi.pi_comparison_plot(self.test_ts)

        _ = epi.get_projection(step=20)
        with self.assertRaises(ValueError):
            # not test_ts provided
            epi.pi_comparison_plot(test_data_only=True)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This module reads synthetic data from daiquery and preprocess it as data_x=features, data_y=hpt_res for given algorithm
"""

import io
import os
import pkgutil
from typing import Any
from typing import Dict
from typing import Optional
from typing import Set

import pandas as pd


class SynthMetadataReader:
    NUM_SECS_IN_DAY: int = 3600 * 24
    PARAMS_TO_SCALE_DOWN: Set[str] = {
        "n_control",
        "n_test",
        "historical_window",
        "scan_window",
    }
    _rawdata: Optional[pd.DataFrame] = None
    _metadata: Optional[Dict[str, Any]] = None

    def _get_raw_data(self) -> pd.DataFrame:
        rawdata = self._rawdata
        if rawdata is None:
            root = "kats"
            path = "data/" if "kats" in os.getcwd().lower() else "kats/data/"
            filename = "meta_learning_detection_training_data_pmo_sample_1000.csv"
            data_object = pkgutil.get_data(root, path + filename)
            assert data_object is not None
            self._rawdata = rawdata = pd.read_csv(
                io.BytesIO(data_object),
                index_col=0,
                dtype={"idx": object},
                encoding="utf8",
            )
        return rawdata.copy()

    def get_metadata(self, algorithm_name: str) -> Dict[str, pd.DataFrame]:
        metadata = self._metadata
        if metadata is None:
            rawdata = self._get_raw_data()

            metadata = {}
            metadata["data_x"] = (
                rawdata.features.map(eval)
                .map(lambda d: {k: float(v) for k, v in d.items()})
                .apply(pd.Series)  # expend dict to columns
            )
            algorithm_names = rawdata.hpt_res.map(eval).map(lambda kv: list(kv.keys())).explode().unique().tolist()

            metadata["data_y"] = {}
            for a in algorithm_names:
                metadata["data_y"][a] = (
                    rawdata.hpt_res.map(eval)
                    .map(lambda kv: kv[a][0])
                    .map(
                        lambda kv: {
                            k: v if k not in self.PARAMS_TO_SCALE_DOWN else v / SynthMetadataReader.NUM_SECS_IN_DAY
                            for k, v in kv.items()
                        }
                    )
                    .apply(pd.Series)  # expend dict to columns
                    .convert_dtypes(convert_integer=False)
                )

            self._metadata = metadata
        return {
            "data_x": metadata["data_x"].copy(),
            "data_y": metadata["data_y"][algorithm_name].copy(),
        }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import importlib
from operator import attrgetter
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import cast
from unittest import TestCase
from unittest.mock import patch

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

import kats.tsfeatures.tsfeatures
from kats.compat import statsmodels
from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.tsfeatures.tsfeatures import _FEATURE_GROUP_MAPPING
from kats.tsfeatures.tsfeatures import TsFeatures


def _univariate_features(feats: Union[Dict[str, float], List[Dict[str, float]]]) -> Dict[str, float]:
    return cast(Dict[str, float], feats)


def _multivariate_features(feats: Union[Dict[str, float], List[Dict[str, float]]]) -> List[Dict[str, float]]:
    return cast(List[Dict[str, float]], feats)


class TSfeaturesTest(TestCase):
    def setUp(self) -> None:
        DATA = load_air_passengers(return_ts=False)
        self.TSData = TimeSeriesData(DATA)
        self.TSData_multivariate = TimeSeriesData(
            pd.DataFrame({"time": DATA.time, "value1": DATA.y, "value2": DATA.y})
        )

        self.TSData_short = TimeSeriesData(DATA.iloc[:8, :])
        self.TSData_mini = TimeSeriesData(DATA.iloc[:2, :])

    def assertDictAlmostEqual(self, expected: Dict[str, Any], features: Dict[str, Any], places: int = 4) -> None:
        """Compares that two dictionaries are floating-point almost equal.

        Note: the dictionaries may or may not contain floating-point values.

        Args:
          expected: the expected dictionary of values.
          actual: the actual dictionary of values.
          places: the number of decimal places for floating point comparisons.
        """

        self.assertEqual(expected.keys(), features.keys())
        for k, v in expected.items():
            if isinstance(v, float):
                if np.isnan(v):
                    self.assertTrue(np.isnan(features[k]), msg=f"{k} differ")
                else:
                    self.assertAlmostEqual(v, features[k], places=4, msg=f"{k} differ")
            else:
                self.assertEqual(v, features[k], msg=f"{k} differ")

    def test_all_feature_names_unique(self) -> None:
        features = set()
        for _, feats in _FEATURE_GROUP_MAPPING.items():
            for feat in feats:
                self.assertFalse(feat in features, f"duplicate feature name {feat}")
                features.add(feat)

    def test_tsfeatures(self) -> None:
        feature_vector = _univariate_features(TsFeatures().transform(self.TSData))

        feature_vector_round = {key: round(feature_vector[key], 6) for key in feature_vector}

        # test there is no nan in feature vector
        self.assertEqual(
            np.isnan(np.asarray(list(feature_vector.values()))).any(),
            False,
        )

        # test there are 40 features in the feature vector now
        self.assertEqual(
            len(np.asarray(list(feature_vector.values()))) == 40,
            True,
        )

        # test feature vector value
        rounded_truth = {
            "length": 144,
            "mean": 280.298611,
            "var": 14291.973331,
            "entropy": 0.428737,
            "lumpiness": 3041164.562906,
            "stability": 12303.627267,
            "flat_spots": 2,
            "hurst": -0.080233,
            "std1st_der": 27.206288,
            "crossing_points": 7,
            "binarize_mean": 0.444444,
            "unitroot_kpss": 0.128475,
            "heterogeneity": 126.064506,
            "histogram_mode": 155.8,
            "linearity": 0.853638,
            "trend_strength": 0.9681,
            "seasonality_strength": 0.440863,
            "spikiness": 33.502886,
            "peak": 6,
            "trough": 3,
            "level_shift_idx": 118,
            "level_shift_size": 15.6,
            "y_acf1": 0.948047,
            "y_acf5": 3.392072,
            "diff1y_acf1": 0.302855,
            "diff1y_acf5": 0.259459,
            "diff2y_acf1": -0.191006,
            "diff2y_acf5": 0.134207,
            "y_pacf5": 1.003288,
            "diff1y_pacf5": 0.219412,
            "diff2y_pacf5": 0.26101,
            "seas_acf1": 0.662904,
            "seas_pacf1": 0.15617,
            "firstmin_ac": 8,
            "firstzero_ac": 52,
            "holt_alpha": 1.0,
            "holt_beta": 0.0,
            "hw_alpha": 0.842106,
            "hw_beta": 0.052631,
            "hw_gamma": 0.157901,
        }
        if statsmodels.version >= "0.12":
            rounded_truth["trend_strength"] = 0.93833
            rounded_truth["seasonality_strength"] = 0.329934
            rounded_truth["spikiness"] = 111.697325
            feature_vector_round["holt_alpha"] = np.round(feature_vector_round["holt_alpha"], 1)
            feature_vector_round["holt_beta"] = np.round(feature_vector_round["holt_beta"], 1)
            rounded_truth["holt_alpha"] = 1.0
            rounded_truth["holt_beta"] = 0.0
            rounded_truth["hw_alpha"] = 1.0
            rounded_truth["hw_beta"] = 0.0
            rounded_truth["hw_gamma"] = 0.0
        self.assertEqual(feature_vector_round, rounded_truth)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand([("univariate", "TSData"), ("multivariate", "TSData_multivariate")])
    def test_feature_selections(self, test_name: str, ts_name: str) -> None:
        # test disabling functions

        ts = attrgetter(ts_name)(self)
        features = TsFeatures(
            unitroot_kpss=False,
            histogram_mode=False,
            diff2y_pacf5=False,
            firstmin_ac=False,
        ).transform(ts)

        if ts.is_univariate():
            feature_list = [_univariate_features(features)]
        else:
            feature_list = _multivariate_features(features)

        rounded_truth = {
            "length": 144,
            "mean": 280.298611,
            "var": 14291.973331,
            "entropy": 0.428737,
            "lumpiness": 3041164.562906,
            "stability": 12303.627267,
            "flat_spots": 2,
            "hurst": -0.080233,
            "std1st_der": 27.206288,
            "crossing_points": 7,
            "binarize_mean": 0.444444,
            "heterogeneity": 126.064506,
            "linearity": 0.853638,
            "trend_strength": 0.9681,
            "seasonality_strength": 0.440863,
            "spikiness": 33.502886,
            "peak": 6,
            "trough": 3,
            "level_shift_idx": 118,
            "level_shift_size": 15.6,
            "y_acf1": 0.948047,
            "y_acf5": 3.392072,
            "diff1y_acf1": 0.302855,
            "diff1y_acf5": 0.259459,
            "diff2y_acf1": -0.191006,
            "diff2y_acf5": 0.134207,
            "y_pacf5": 1.003288,
            "diff1y_pacf5": 0.219412,
            "seas_acf1": 0.662904,
            "seas_pacf1": 0.15617,
            "firstzero_ac": 52,
            "holt_alpha": 1.0,
            "holt_beta": 0.0,
            "hw_alpha": 0.842106,
            "hw_beta": 0.052631,
            "hw_gamma": 0.157901,
        }

        for feature_vector in feature_list:
            if statsmodels.version >= "0.12":
                rounded_truth["trend_strength"] = 0.93833
                rounded_truth["seasonality_strength"] = 0.329934
                rounded_truth["spikiness"] = 111.697325
                feature_vector["holt_alpha"] = np.round(feature_vector["holt_alpha"], 1)
                feature_vector["holt_beta"] = np.round(feature_vector["holt_beta"], 1)
                rounded_truth["holt_alpha"] = 1.0
                rounded_truth["holt_beta"] = 0.0
                rounded_truth["hw_alpha"] = 1.0
                rounded_truth["hw_beta"] = 0.0
                rounded_truth["hw_gamma"] = 0.0
            self.assertDictAlmostEqual(rounded_truth, feature_vector, places=6)

        # test selecting features
        features = [
            "var",
            "linearity",
            "spikiness",
            "trough",
            "holt_alpha",
            "hw_gamma",
            "level_shift_idx",
        ]

        features = TsFeatures(selected_features=features).transform(ts)

        if ts.is_univariate():
            feature_list = [_univariate_features(features)]
        else:
            feature_list = _multivariate_features(features)

        # test feature vector value
        rounded_truth = {
            "var": 14291.973331,
            "linearity": 0.853638,
            "spikiness": 33.502886,
            "trough": 3,
            "level_shift_idx": 118,
            "holt_alpha": 1.0,
            "hw_gamma": 0.157901,
        }

        for feature_vector in feature_list:
            if statsmodels.version >= "0.12":
                rounded_truth["spikiness"] = 111.697325
                feature_vector["holt_alpha"] = np.round(feature_vector["holt_alpha"], 1)
                rounded_truth["holt_alpha"] = 1.0
                rounded_truth["hw_gamma"] = 0.0

            self.assertDictAlmostEqual(rounded_truth, feature_vector, places=6)

        # test selecting extension features
        extension_features = [
            "cusum_num",
            "cusum_conf",
            "cusum_cp_index",
            "cusum_delta",
            "cusum_llr",
            "cusum_regression_detected",
            "cusum_stable_changepoint",
            "cusum_p_value",
            "robust_num",
            "robust_metric_mean",
            "outlier_num",
            "trend_num",
            "trend_num_increasing",
            "trend_avg_abs_tau",
            "nowcast_roc",
            "nowcast_ma",
            "nowcast_mom",
            "nowcast_lag",
            "nowcast_macd",
            "nowcast_macdsign",
            "nowcast_macddiff",
            "seasonal_period",
            "seasonality_mag",
            "trend_mag",
            "residual_std",
        ]

        features = TsFeatures(selected_features=extension_features).transform(ts)

        if ts.is_univariate():
            feature_list = [_univariate_features(features)]
        else:
            feature_list = _multivariate_features(features)

        # test feature vector value
        rounded_truth = {
            "cusum_num": 1,
            "cusum_conf": 1.0,
            "cusum_cp_index": 0.527778,
            "cusum_delta": 199.098856,
            "cusum_llr": 168.663483,
            "cusum_regression_detected": 1,
            "cusum_stable_changepoint": 1,
            "cusum_p_value": 0.0,
            "robust_num": 3,
            "robust_metric_mean": -31.866667,
            "outlier_num": 0,
            "trend_num": 2,
            "trend_num_increasing": 2,
            "trend_avg_abs_tau": 0.821053,
            "nowcast_roc": 0.062858,
            "nowcast_ma": 280.417143,
            "nowcast_mom": 12.841727,
            "nowcast_lag": 273.136691,
            "nowcast_macd": 11.032608,
            "nowcast_macdsign": 10.985509,
            "nowcast_macddiff": 0.527714,
            "seasonal_period": 7,
            "trend_mag": 2.404464,
            "seasonality_mag": 35.0,
            "residual_std": 21.258429,
        }
        for feature_vector in feature_list:
            if statsmodels.version >= "0.12":
                rounded_truth["trend_mag"] = 2.318814
                rounded_truth["seasonality_mag"] = 36.0
                rounded_truth["residual_std"] = 29.630087
            self.assertDictAlmostEqual(rounded_truth, feature_vector, places=6)

    def test_others(self) -> None:
        # test there is nan in feature vector because the length of TS is too short
        feature_vector = _univariate_features(TsFeatures().transform(self.TSData_short))

        self.assertEqual(
            np.isnan(np.asarray(list(feature_vector.values()))).any(),
            True,
        )

    def test_errors(self) -> None:
        # test input error (time series is too short)
        self.assertRaises(
            ValueError,
            TsFeatures().transform,
            self.TSData_mini,
        )
        with self.assertRaises(ValueError):
            TsFeatures(selected_features=["mango"])

    def test_acfpacf_warning(self) -> None:
        value = np.arange(30)
        time_ = pd.date_range(end="2022-08-01", freq="1D", periods=30)
        tsd = TimeSeriesData(pd.DataFrame({"value": value, "time": time_}))
        features = _univariate_features(TsFeatures(selected_features=["acfpacf_features"]).transform(tsd))
        acfpacf_features = [features[f] for f in _FEATURE_GROUP_MAPPING["acfpacf_features"]]
        self.assertEqual(acfpacf_features, [np.nan] * len(acfpacf_features))

    def test_IntegerArrays(self) -> None:
        if statsmodels.version < "0.12":
            df = pd.DataFrame(
                {
                    "time": range(15),
                    "value": [1, 4, 9, 4, 5, 5, 7, 2, 5, 1, 6, 3, 6, 5, 5],
                }
            )
        elif statsmodels.version >= "0.12":
            df = pd.DataFrame(
                {
                    "time": range(20),
                    "value": [
                        1,
                        4,
                        9,
                        4,
                        5,
                        5,
                        7,
                        2,
                        5,
                        1,
                        6,
                        3,
                        6,
                        5,
                        5,
                        6,
                        9,
                        10,
                        5,
                        6,
                    ],
                }
            )
        df["value"] = df["value"].astype(dtype=pd.Int64Dtype())
        # pyre-fixme[61]: `df` may not be initialized here.
        ts = TimeSeriesData(df)

        ts_features = TsFeatures(
            selected_features=[
                "length",
                "mean",
                "entropy",
                "hurst",
                "y_acf1",
                "seas_acf1",
                "hw_gamma",
            ]
        )
        feats = _univariate_features(ts_features.transform(ts))
        feats = {key: round(feats[key], 3) for key in feats}
        if statsmodels.version < "0.12":
            self.assertEqual(
                {
                    "length": 15,
                    "mean": 4.533,
                    "entropy": 0.765,
                    "hurst": -0.143,
                    "y_acf1": -0.298,
                    "seas_acf1": -0.121,
                    "hw_gamma": 0.947,
                },
                feats,
            )
        elif statsmodels.version >= "0.12":
            self.assertEqual(
                {
                    "length": 20,
                    "mean": 5.2,
                    "entropy": 0.894,
                    "hurst": -0.12,
                    "y_acf1": 0.041,
                    "seas_acf1": -0.125,
                    "hw_gamma": 0.0,
                },
                feats,
            )

    def test_tsfeatures_time(self) -> None:
        dates = pd.date_range("2021-01-22", "2021-01-31", tz="US/Pacific").tolist()
        dates += pd.date_range("2022-01-30", "2022-02-04", tz="US/Pacific").tolist()
        ts = TimeSeriesData(
            df=pd.DataFrame(
                {
                    "time": dates,
                    "value": [10, 2, 11, 5, 13, 18, 4, 14, 17, 9] + [np.nan] * 6,
                }
            )
        )
        features = _univariate_features(TsFeatures(selected_features=["time"]).transform(ts))
        expected = {
            "time_years": 2,
            "time_months": 3,
            "time_monthsofyear": 2,
            "time_weeks": 4,
            "time_weeksofyear": 3,
            "time_days": 16,
            "time_daysofyear": 14,
            "time_avg_timezone_offset": -28800.0,
            "time_length_days": 378,
            "time_freq_Monday": 0.125,
            "time_freq_Tuesday": 0.125,
            "time_freq_Wednesday": 0.125,
            "time_freq_Thursday": 0.125,
            "time_freq_Friday": 0.1875,
            "time_freq_Saturday": 0.125,
            "time_freq_Sunday": 0.1875,
        }
        self.assertDictAlmostEqual(expected, features)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""CUSUMDetectorModel is a wraper of CUSUMDetector to detect multiple change points

Typical usage example:

>>> # Define CUSUMDetectorModel
>>> model = CUSUMDetectorModel(
        scan_window=43200,
        historical_window=604800,
        threshold=0.01,
        delta_std_ratio=1.0,
        serialized_model=None,
        change_directions=["increase"],
        score_func=CusumScoreFunction.percentage_change,
        remove_seasonality=True,
    )
>>> # Run detector
>>> respond = model.fit_predict(tsd)
>>> # Plot anomaly score
>>> respond.scores.plot(cols=['value'])
>>> # Get change points in unixtime
>>> change_points = model.cps
"""

import json
import logging
from datetime import datetime
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Union
from typing import cast

import numpy as np
import pandas as pd

from kats.consts import DEFAULT_VALUE_NAME
from kats.consts import IRREGULAR_GRANULARITY_ERROR
from kats.consts import TimeSeriesData
from kats.detectors.cusum_detection import CUSUMDefaultArgs
from kats.detectors.cusum_detection import CUSUMDetector
from kats.detectors.detector import DetectorModel
from kats.detectors.detector_consts import AnomalyResponse
from kats.utils.decomposition import TimeSeriesDecomposition

NORMAL_TOLERENCE = 1  # number of window
CHANGEPOINT_RETENTION: int = 7 * 24 * 60 * 60  # in seconds
MAX_CHANGEPOINT = 10
SEASON_PERIOD_FREQ_MAP: Dict[str, int] = {
    "daily": 1,
    "weekly": 7,
    "monthly": 30,
    "yearly": 365,
}

_log: logging.Logger = logging.getLogger("cusum_model")


def percentage_change(data: TimeSeriesData, pre_mean: float, **kwargs: Any) -> TimeSeriesData:
    """
    Calculate percentage change absolute change / baseline change

    Args:
        data: The data need to calculate the score
        pre_mean: Baseline mean
    """

    return (data - pre_mean) / (pre_mean)


def change(data: TimeSeriesData, pre_mean: float, **kwargs: Any) -> TimeSeriesData:
    """
    Calculate absolute change

    Args:
        data: The data need to calculate the score
        pre_mean: Baseline mean
    """

    return data - pre_mean


def z_score(data: TimeSeriesData, pre_mean: float, pre_std: float) -> TimeSeriesData:
    """
    Calculate z score: absolute change / std

    Args:
        data: The data need to calculate the score
        pre_mean: Baseline mean
        pre_std: Baseline std
    """

    return (data - pre_mean) / (pre_std)


class CusumScoreFunction(Enum):
    change = "change"
    percentage_change = "percentage_change"
    z_score = "z_score"


# Score Function Constants
# pyre-fixme
SCORE_FUNC_DICT = {
    CusumScoreFunction.change.value: change,
    CusumScoreFunction.percentage_change.value: percentage_change,
    CusumScoreFunction.z_score.value: z_score,
}
DEFAULT_SCORE_FUNCTION: CusumScoreFunction = CusumScoreFunction.change
STR_TO_SCORE_FUNC: Dict[str, CusumScoreFunction] = {  # Used for param tuning
    "change": CusumScoreFunction.change,
    "percentage_change": CusumScoreFunction.percentage_change,
    "z_score": CusumScoreFunction.z_score,
}


class PredictFunctionValues(NamedTuple):
    score: TimeSeriesData
    absolute_change: TimeSeriesData


class CUSUMDetectorModel(DetectorModel):
    """CUSUMDetectorModel for detecting multiple level shift change points

    CUSUMDetectorModel runs CUSUMDetector multiple times to detect multiple change
    points. In each run, CUSUMDetector will use historical_window + scan_window as
    input time series, and find change point in scan_window. The DetectorModel stores
    change points and returns anomaly score.

    Attributes:
        cps: Change points detected in unixtime.
        alert_fired: If a change point is detected and the anomaly still present.
        pre_mean: Previous baseline mean.
        pre_std: Previous baseline std.
        number_of_normal_scan: Number of scans with mean returned back to baseline.
        alert_change_direction: Increase or decrease.
        scan_window: Length in seconds of scan window.
        historical_window: Length in seconds of historical window.
        step_window: The time difference between CUSUM runs.
        threshold: CUSUMDetector threshold.
        delta_std_ratio: The mean delta have to larger than this parameter times std of
            the data to be consider as a change.
        magnitude_quantile: See in CUSUMDetector.
        magnitude_ratio: See in CUSUMDetector.
        score_func: The score function to calculate the anomaly score.
        remove_seasonality: If apply STL to remove seasonality.
        season_period_freq: str, "daily"/"weekly"/"monthly"/"yearly"
    """

    def __init__(
        self,
        serialized_model: Optional[bytes] = None,
        scan_window: Optional[int] = None,
        historical_window: Optional[int] = None,
        step_window: Optional[int] = None,
        threshold: float = CUSUMDefaultArgs.threshold,
        delta_std_ratio: float = CUSUMDefaultArgs.delta_std_ratio,
        magnitude_quantile: Optional[float] = CUSUMDefaultArgs.magnitude_quantile,
        magnitude_ratio: float = CUSUMDefaultArgs.magnitude_ratio,
        change_directions: Optional[Union[List[str], str]] = CUSUMDefaultArgs.change_directions,
        score_func: Union[str, CusumScoreFunction] = DEFAULT_SCORE_FUNCTION,
        remove_seasonality: bool = CUSUMDefaultArgs.remove_seasonality,
        season_period_freq: str = "daily",
    ) -> None:
        if serialized_model:
            previous_model = json.loads(serialized_model)
            self.cps: List[int] = previous_model["cps"]
            self.alert_fired: bool = previous_model["alert_fired"]
            self.pre_mean: float = previous_model["pre_mean"]
            self.pre_std: float = previous_model["pre_std"]
            self.number_of_normal_scan: int = previous_model["number_of_normal_scan"]
            self.alert_change_direction: str = previous_model["alert_change_direction"]
            self.scan_window: int = previous_model["scan_window"]
            scan_window = previous_model["scan_window"]
            self.historical_window: int = previous_model["historical_window"]
            self.step_window: int = previous_model["step_window"]
            step_window = previous_model["step_window"]
            self.threshold: float = previous_model["threshold"]
            self.delta_std_ratio: float = previous_model["delta_std_ratio"]
            self.magnitude_quantile: Optional[float] = previous_model["magnitude_quantile"]
            self.magnitude_ratio: float = previous_model["magnitude_ratio"]
            self.change_directions: Optional[List[str]] = previous_model["change_directions"]
            self.score_func: CusumScoreFunction = previous_model["score_func"]
            if "remove_seasonality" in previous_model:
                self.remove_seasonality: bool = previous_model["remove_seasonality"]
            else:
                self.remove_seasonality: bool = remove_seasonality

            self.season_period_freq: str = previous_model.get("season_period_freq", "daily")

        elif scan_window is not None and historical_window is not None:
            self.cps = []
            self.alert_fired = False
            self.pre_mean = 0
            self.pre_std = 1
            self.number_of_normal_scan = 0
            self.alert_change_direction: Union[str, None] = None
            self.scan_window = scan_window
            self.historical_window = historical_window
            self.step_window = cast(int, step_window)
            self.threshold = threshold
            self.delta_std_ratio = delta_std_ratio
            self.magnitude_quantile = magnitude_quantile
            self.magnitude_ratio = magnitude_ratio

            if isinstance(change_directions, str):
                self.change_directions = [change_directions]
            else:
                # List[str]
                self.change_directions = change_directions

            self.remove_seasonality = remove_seasonality
            self.season_period_freq = season_period_freq

            # We allow score_function to be a str for compatibility with param tuning
            if isinstance(score_func, str):
                if score_func in STR_TO_SCORE_FUNC:
                    score_func = STR_TO_SCORE_FUNC[score_func]
                else:
                    score_func = DEFAULT_SCORE_FUNCTION
            self.score_func = score_func.value

        else:
            raise ValueError(
                "You must provide either serialized model or values for " "scan_window and historical_window."
            )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CUSUMDetectorModel):
            return (
                self.cps == other.cps
                and self.alert_fired == other.alert_fired
                and self.pre_mean == other.pre_mean
                and self.pre_std == other.pre_std
                and self.number_of_normal_scan == other.number_of_normal_scan
                and self.alert_change_direction == other.alert_change_direction
                and self.scan_window == other.scan_window
                and self.historical_window == other.historical_window
                and self.step_window == other.step_window
                and self.threshold == other.threshold
                and self.delta_std_ratio == other.delta_std_ratio
                and self.magnitude_quantile == other.magnitude_quantile
                and self.magnitude_ratio == other.magnitude_ratio
                and self.change_directions == other.change_directions
                and self.score_func == other.score_func
            )
        return False

    def serialize(self) -> bytes:
        """
        Retrun serilized model.
        """

        return str.encode(json.dumps(self.__dict__))

    def _set_alert_off(self) -> None:
        self.alert_fired = False
        self.number_of_normal_scan = 0

    def _set_alert_on(self, baseline_mean: float, baseline_std: float, alert_change_direction: str) -> None:
        self.alert_fired = True
        self.alert_change_direction = alert_change_direction
        self.pre_mean = baseline_mean
        self.pre_std = baseline_std

    def _if_normal(self, cur_mean: float, change_directions: Optional[List[str]]) -> bool:
        if change_directions is not None:
            increase, decrease = (
                "increase" in change_directions,
                "decrease" in change_directions,
            )
        else:
            increase, decrease = True, True

        if self.alert_change_direction == "increase":
            check_increase = 0 if increase else np.inf
            check_decrease = 1.0 if decrease else np.inf
        elif self.alert_change_direction == "decrease":
            check_increase = 1.0 if increase else np.inf
            check_decrease = 0 if decrease else np.inf

        return (
            self.pre_mean - check_decrease * self.pre_std <= cur_mean <= self.pre_mean + check_increase * self.pre_std
        )

    def _fit(
        self,
        data: TimeSeriesData,
        historical_data: TimeSeriesData,
        scan_window: Union[int, pd.Timedelta],
        threshold: float = CUSUMDefaultArgs.threshold,
        delta_std_ratio: float = CUSUMDefaultArgs.delta_std_ratio,
        magnitude_quantile: Optional[float] = CUSUMDefaultArgs.magnitude_quantile,
        magnitude_ratio: float = CUSUMDefaultArgs.magnitude_ratio,
        change_directions: Optional[List[str]] = CUSUMDefaultArgs.change_directions,
    ) -> None:
        """Fit CUSUM model.

        Args:
            data: the new data the model never seen
            historical_data: the historical data, `historical_data` have to end with the
                datapoint right before the first data point in `data`
            scan_window: scan window length in seconds, scan window is the window where
                cusum search for changepoint(s)
            threshold: changepoint significant level, higher the value more changepoints
                detected
            delta_std_ratio: the mean change have to larger than `delta_std_ratio` *
            `std(data[:changepoint])` to be consider as a change, higher the value
            less changepoints detected
            magnitude_quantile: float, the quantile for magnitude comparison, if
                none, will skip the magnitude comparison;
            magnitude_ratio: float, comparable ratio;
            change_directions: a list contain either or both 'increase' and 'decrease' to
                specify what type of change to detect;
        """
        historical_data.extend(data, validate=False)
        n = len(historical_data)
        scan_start_time = historical_data.time.iloc[-1] - pd.Timedelta(scan_window, unit="s")
        scan_start_index = max(0, np.argwhere((historical_data.time >= scan_start_time).values).min())
        if not self.alert_fired:
            # if scan window is less than 2 data poins and there is no alert fired
            # skip this scan
            if n - scan_start_index <= 1:
                return
            detector = CUSUMDetector(historical_data)
            changepoints = detector.detector(
                interest_window=[scan_start_index, n],
                threshold=threshold,
                delta_std_ratio=delta_std_ratio,
                magnitude_quantile=magnitude_quantile,
                magnitude_ratio=magnitude_ratio,
                change_directions=change_directions,
            )
            if len(changepoints) > 0:
                cp = sorted(changepoints, key=lambda x: x.start_time)[0]
                self.cps.append(int(cp.start_time.value / 1e9))

                if len(self.cps) > MAX_CHANGEPOINT:
                    self.cps.pop(0)

                self._set_alert_on(
                    historical_data.value[: cp.cp_index + 1].mean(),
                    historical_data.value[: cp.cp_index + 1].std(),
                    cp.direction,
                )
        else:
            cur_mean = historical_data[scan_start_index:].value.mean()

            if self._if_normal(cur_mean, change_directions):
                self.number_of_normal_scan += 1
                if self.number_of_normal_scan >= NORMAL_TOLERENCE:
                    self._set_alert_off()
            else:
                self.number_of_normal_scan = 0

            current_time = int(data.time.max().value / 1e9)
            if current_time - self.cps[-1] > CHANGEPOINT_RETENTION:
                self._set_alert_off()

    def _predict(
        self,
        data: TimeSeriesData,
        score_func: CusumScoreFunction = CusumScoreFunction.change,
    ) -> PredictFunctionValues:
        """
        data: the new data for the anoamly score calculation.
        """
        if self.alert_fired:
            cp = self.cps[-1]
            tz = data.tz()
            if tz is None:
                change_time = pd.to_datetime(cp, unit="s")
            else:
                change_time = pd.to_datetime(cp, unit="s", utc=True).tz_convert(tz)

            if change_time >= data.time.iloc[0]:
                cp_index = data.time[data.time == change_time].index[0]
                data_pre = data[: cp_index + 1]
                score_pre = self._zeros_ts(data_pre)
                change_pre = self._zeros_ts(data_pre)
                score_post = SCORE_FUNC_DICT[score_func](
                    data=data[cp_index + 1 :],
                    pre_mean=self.pre_mean,
                    pre_std=self.pre_std,
                )
                score_pre.extend(score_post, validate=False)

                change_post = SCORE_FUNC_DICT[CusumScoreFunction.change.value](
                    data=data[cp_index + 1 :],
                    pre_mean=self.pre_mean,
                    pre_std=self.pre_std,
                )
                change_pre.extend(change_post, validate=False)
                return PredictFunctionValues(score_pre, change_pre)
            return PredictFunctionValues(
                SCORE_FUNC_DICT[score_func](data=data, pre_mean=self.pre_mean, pre_std=self.pre_std),
                SCORE_FUNC_DICT[CusumScoreFunction.change.value](
                    data=data, pre_mean=self.pre_mean, pre_std=self.pre_std
                ),
            )
        else:
            return PredictFunctionValues(self._zeros_ts(data), self._zeros_ts(data))

    def _zeros_ts(self, data: TimeSeriesData) -> TimeSeriesData:
        return TimeSeriesData(
            time=data.time,
            value=pd.Series(
                np.zeros(len(data)),
                name=data.value.name if data.value.name else DEFAULT_VALUE_NAME,
                copy=False,
            ),
        )

    def _check_window_sizes(self, frequency_sec: int) -> None:
        """
        Function to check if historical_window, scan_window, and step_window
        are suitable for a given TS data and a given TS historical_data.
        We have already checked if self.step_window < self.scan_window in init func
        when self.step_window is not None.
        """
        if self.step_window is not None:
            if self.step_window >= self.scan_window:
                raise ValueError(
                    "Step window is supposed to be smaller than scan window to ensure we "
                    "have overlap for scan windows."
                )
            if self.step_window < frequency_sec:
                raise ValueError(
                    "Step window is supposed to be greater than TS granularity. "
                    f"TS granularity is: {frequency_sec} seconds. "
                    "Please provide a larger step window."
                )

        # if step_window is None, step_window = min(self.scan_window/2, *** + frequency_sec)
        # in order to make sure step_window >= frequency_sec, we need to make sure
        # self.scan_window >= 2 * frequency_sec

        if self.scan_window < 2 * frequency_sec or self.historical_window < 2 * frequency_sec:
            raise ValueError(
                "Scan window and historical window are supposed to be >= 2 * TS granularity. "
                f"TS granularity is: {frequency_sec} seconds. "
                "Please provide a larger scan window or historical_window."
            )

    def fit_predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """
        This function combines fit and predict and return anomaly socre for data. It
        requires scan_window > step_window.
        The relationship between two consective cusum runs in the loop is shown as below:

        >>> |---historical_window---|---scan_window---|
        >>>                                           |-step_window-|
        >>>               |---historical_window---|---scan_window---|

        # requirement: scan window > step window
        * scan_window: the window size in seconds to detect change point
        * historical_window: the window size in seconds to provide historical data
        * step_window: the window size in seconds to specify the step size between two scans

        Args:
            data: :class:`kats.consts.TimeSeriesData` object representing the data
            historical_data: :class:`kats.consts.TimeSeriesData` object representing the history.

        Returns:
            The anomaly response contains the anomaly scores.
        """
        # get parameters
        scan_window = self.scan_window
        historical_window = self.historical_window
        step_window = self.step_window
        threshold = self.threshold
        delta_std_ratio = self.delta_std_ratio
        magnitude_quantile = self.magnitude_quantile
        magnitude_ratio = self.magnitude_ratio
        change_directions = self.change_directions
        score_func = self.score_func
        remove_seasonality = self.remove_seasonality
        season_period_freq = self.season_period_freq

        scan_window = pd.Timedelta(scan_window, unit="s")
        historical_window = pd.Timedelta(historical_window, unit="s")

        # pull all the data in historical data
        if historical_data is not None:
            # make a copy of historical data
            historical_data = historical_data[:]
            historical_data.extend(data, validate=False)
        else:
            # When historical_data is not provided, will use part of data as
            # historical_data, and fill with zero anomaly score.
            historical_data = data[:]

        frequency = historical_data.freq_to_timedelta()
        if frequency is None or frequency is pd.NaT:
            # Use the top frequency if any, when not able to infer from data.
            freq_counts = historical_data.time.diff().value_counts().sort_values(ascending=False)
            if freq_counts.iloc[0] >= int(len(historical_data)) * 0.8 - 1:
                frequency = freq_counts.index[0]
            else:
                _log.debug(f"freq_counts: {freq_counts}")
                raise ValueError(IRREGULAR_GRANULARITY_ERROR)

        # check if historical_window, scan_window, and step_window are suitable for given TSs
        self._check_window_sizes(frequency.total_seconds())

        if remove_seasonality:
            frequency_sec: int = int(frequency.total_seconds())
            frequency_sec_str = str(frequency_sec) + "s"

            # calculate resample base in second level
            # calculate remainder as resampling base
            resample_base_sec = (
                pd.to_datetime(historical_data.time[0]).day * 24 * 60 * 60
                + pd.to_datetime(historical_data.time[0]).hour * 60 * 60
                + pd.to_datetime(historical_data.time[0]).minute * 60
                + pd.to_datetime(historical_data.time[0]).second
            ) % frequency_sec

            decomposer_input = historical_data.interpolate(
                freq=frequency_sec_str,
                base=resample_base_sec,
            )

            data_time_idx = decomposer_input.time.isin(historical_data.time)
            if len(decomposer_input.time[data_time_idx]) != len(historical_data):
                raise ValueError(IRREGULAR_GRANULARITY_ERROR)

            # fixing the period to 24 hours as indicated in T81530775
            season_period_freq_multiplier = SEASON_PERIOD_FREQ_MAP[season_period_freq]
            period = int(24 * 60 * 60 * season_period_freq_multiplier / frequency.total_seconds())

            if period < 2:
                period = 7

            decomposer = TimeSeriesDecomposition(
                decomposer_input,
                # statsmodels.STL requires that period must be a positive integer >= 2
                period=max(period, 2),
                robust=True,
                seasonal_deg=0,
                trend_deg=1,
                low_pass_deg=1,
                # statsmodels.STL requires that low_pass_jump must be a positive integer
                low_pass_jump=max(int((period + 1) * 0.15), 1),  # save run time
                seasonal_jump=1,
                # statsmodels.STL requires that trend_jump must be a positive integer
                trend_jump=max(int((period + 1) * 0.15), 1),  # save run time
            )

            decomp = decomposer.decomposer()
            historical_data_time_idx = decomp["rem"].time.isin(historical_data.time)

            historical_data.value = pd.Series(
                decomp["rem"][historical_data_time_idx].value + decomp["trend"][historical_data_time_idx].value,
                name=historical_data.value.name,
                copy=False,
            )

        smooth_window = int(scan_window.total_seconds() / frequency.total_seconds())
        if smooth_window > 1:
            smooth_historical_value = pd.Series(
                np.convolve(historical_data.value.values, np.ones(smooth_window), mode="full")[: 1 - smooth_window]
                / smooth_window,
                name=historical_data.value.name,
                copy=False,
            )
            smooth_historical_data = TimeSeriesData(time=historical_data.time, value=smooth_historical_value)
        else:
            smooth_historical_data = historical_data

        anomaly_start_time = max(historical_data.time.iloc[0] + historical_window, data.time.iloc[0])
        if anomaly_start_time > historical_data.time.iloc[-1]:
            # if len(all data) is smaller than historical window return zero score
            # Calling first _predict to poulate self.change_point_delta
            predict_results = self._predict(smooth_historical_data[-len(data) :], score_func)
            return AnomalyResponse(
                scores=predict_results.score,
                confidence_band=None,
                predicted_ts=None,
                anomaly_magnitude_ts=predict_results.absolute_change,
                stat_sig_ts=None,
            )
        anomaly_start_idx = self._time2idx(data, anomaly_start_time, "right")
        anomaly_start_time = data.time.iloc[anomaly_start_idx]
        score_tsd = self._zeros_ts(data[:anomaly_start_idx])
        change_tsd = self._zeros_ts(data[:anomaly_start_idx])

        if historical_data.time.iloc[-1] - historical_data.time.iloc[0] + frequency <= scan_window:
            # if len(all data) is smaller than scan data return zero score
            # Calling first _predict to poulate self.change_point_delta
            predict_results = self._predict(smooth_historical_data[-len(data) :], score_func)
            return AnomalyResponse(
                scores=predict_results.score,
                confidence_band=None,
                predicted_ts=None,
                anomaly_magnitude_ts=predict_results.absolute_change,
                stat_sig_ts=None,
            )

        if step_window is None:
            # if step window is not provide use the time range of data or
            # half of the scan_window.
            step_window = min(
                scan_window / 2,
                (data.time.iloc[-1] - data.time.iloc[0]) + frequency,  # to include the last data point
            )
        else:
            step_window = pd.Timedelta(step_window, unit="s")

        for start_time in pd.date_range(
            anomaly_start_time,
            min(
                data.time.iloc[-1] + frequency - step_window,  # to include last data point
                data.time.iloc[-1],  # make sure start_time won't beyond last data time
            ),
            freq=step_window,
        ):
            _log.debug(f"start_time {start_time}")
            historical_start = self._time2idx(historical_data, start_time - historical_window, "right")
            _log.debug(f"historical_start {historical_start}")
            historical_end = self._time2idx(historical_data, start_time, "right")
            _log.debug(f"historical_end {historical_end}")
            scan_end = self._time2idx(historical_data, start_time + step_window, "left")
            _log.debug(f"scan_end {scan_end}")
            in_data = historical_data[historical_end : scan_end + 1]
            if len(in_data) == 0:
                # skip if there is no data in the step_window
                continue
            in_hist = historical_data[historical_start:historical_end]
            self._fit(
                in_data,
                in_hist,
                scan_window=cast(Union[int, pd.Timedelta], scan_window),
                threshold=threshold,
                delta_std_ratio=delta_std_ratio,
                magnitude_quantile=magnitude_quantile,
                magnitude_ratio=magnitude_ratio,
                change_directions=change_directions,
            )
            predict_results = self._predict(
                smooth_historical_data[historical_end : scan_end + 1],
                score_func=score_func,
            )
            score_tsd.extend(
                predict_results.score,
                validate=False,
            )
            change_tsd.extend(predict_results.absolute_change, validate=False)

        # Handle the remaining data
        remain_data_len = len(data) - len(score_tsd)
        if remain_data_len > 0:
            scan_end = len(historical_data)
            historical_end = len(historical_data) - remain_data_len
            historical_start = self._time2idx(
                historical_data,
                historical_data.time.iloc[historical_end] - historical_window,
                "left",
            )
            in_data = historical_data[historical_end:scan_end]
            in_hist = historical_data[historical_start:historical_end]
            self._fit(
                in_data,
                in_hist,
                scan_window=cast(Union[int, pd.Timedelta], scan_window),
                threshold=threshold,
                delta_std_ratio=delta_std_ratio,
                magnitude_quantile=magnitude_quantile,
                magnitude_ratio=magnitude_ratio,
                change_directions=change_directions,
            )
            predict_results = self._predict(
                smooth_historical_data[historical_end:scan_end],
                score_func=score_func,
            )
            score_tsd.extend(
                predict_results.score,
                validate=False,
            )
            change_tsd.extend(predict_results.absolute_change, validate=False)

        return AnomalyResponse(
            scores=score_tsd,
            confidence_band=None,
            predicted_ts=None,
            anomaly_magnitude_ts=change_tsd,
            stat_sig_ts=None,
        )

    def _time2idx(self, tsd: TimeSeriesData, time: datetime, direction: str) -> int:
        """
        This function get the index of the TimeSeries data given a datatime.
        left takes the index on the left of the time stamp (inclusive)
        right takes the index on the right of the time stamp (exclusive)
        """
        if direction == "right":
            return np.argwhere((tsd.time >= time).values).min()
        elif direction == "left":
            return np.argwhere((tsd.time < time).values).max()
        else:
            raise ValueError("direction can only be right or left")

    def fit(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> None:
        self.fit_predict(data, historical_data)

    def predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """
        predict is not implemented
        """
        raise ValueError("predict is not implemented, call fit_predict() instead")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""A module for meta-learner hyper-parameter selection.

This module contains two classes, including:
    - :class:`MetaLearnHPT` for recommending hyper-parameters of forecasting models;
    - :class:`MultitaskNet` for multi-task neural network built with pytorch.
"""

import collections
import logging
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import joblib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from sklearn.model_selection import train_test_split

from kats.consts import TimeSeriesData
from kats.tsfeatures.tsfeatures import TsFeatures

_MODELS = {
    "arima",
    "sarima",
    "holtwinters",
    "stlf",
    "theta",
    "cusum",
    "statsig",
}

logging.basicConfig()
LOGGER: logging.Logger = logging.getLogger(__name__)


@dataclass
class DefaultModelParams:
    holtwinters_categorical_idx: List[str] = field(default_factory=list)
    holtwinters_numerical_idx: List[str] = field(default_factory=list)
    arima_categorical_idx: List[str] = field(default_factory=list)
    arima_numerical_idx: List[str] = field(default_factory=list)
    sarima_categorical_idx: List[str] = field(default_factory=list)
    sarima_numerical_idx: List[str] = field(default_factory=list)
    theta_categorical_idx: List[str] = field(default_factory=list)
    theta_numerical_idx: List[str] = field(default_factory=list)
    stlf_categorical_idx: List[str] = field(default_factory=list)
    stlf_numerical_idx: List[str] = field(default_factory=list)
    cusum_categorical_idx: List[str] = field(default_factory=list)
    cusum_numerical_idx: List[str] = field(default_factory=list)
    statsig_categorical_idx: List[str] = field(default_factory=list)
    statsig_numerical_idx: List[str] = field(default_factory=list)

    def __init__(self) -> None:
        self.holtwinters_categorical_idx = [
            "trend",
            "damped",
            "seasonal",
            "seasonal_periods",
        ]
        self.holtwinters_numerical_idx = []
        self.arima_categorical_idx = ["p", "d", "q"]
        self.arima_numerical_idx = []
        self.sarima_categorical_idx = ["seasonal_order", "trend", "p", "d", "q"]
        self.sarima_numerical_idx = []
        self.theta_categorical_idx = ["m"]
        self.theta_numerical_idx = []
        self.stlf_categorical_idx = ["method", "m"]
        self.stlf_numerical_idx = []
        self.cusum_categorical_idx = ["score_func"]
        self.cusum_numerical_idx = [
            "delta_std_ratio",
            "scan_window",
            "historical_window",
        ]
        self.statsig_categorical_idx = []
        self.statsig_numerical_idx = ["n_control", "n_test"]


@dataclass
class DefaultModelNetworks:
    holtwinters_n_hidden_shared: List[int] = field(default_factory=list)
    holtwinters_n_hidden_cat_combo: List[List[int]] = field(default_factory=list)
    holtwinters_n_hidden_num: List[int] = field(default_factory=list)
    arima_n_hidden_shared: List[int] = field(default_factory=list)
    arima_n_hidden_cat_combo: List[List[int]] = field(default_factory=list)
    arima_n_hidden_num: List[int] = field(default_factory=list)
    sarima_n_hidden_shared: List[int] = field(default_factory=list)
    sarima_n_hidden_cat_combo: List[List[int]] = field(default_factory=list)
    sarima_n_hidden_num: List[int] = field(default_factory=list)
    theta_n_hidden_shared: List[int] = field(default_factory=list)
    theta_n_hidden_cat_combo: List[List[int]] = field(default_factory=list)
    theta_n_hidden_num: List[int] = field(default_factory=list)
    stlf_n_hidden_shared: List[int] = field(default_factory=list)
    stlf_n_hidden_cat_combo: List[List[int]] = field(default_factory=list)
    stlf_n_hidden_num: List[int] = field(default_factory=list)
    cusum_n_hidden_shared: List[int] = field(default_factory=list)
    cusum_n_hidden_cat_combo: List[List[int]] = field(default_factory=list)
    cusum_n_hidden_num: List[int] = field(default_factory=list)
    statsig_n_hidden_shared: List[int] = field(default_factory=list)
    statsig_n_hidden_cat_combo: List[List[int]] = field(default_factory=list)
    statsig_n_hidden_num: List[int] = field(default_factory=list)

    def __init__(self) -> None:
        self.holtwinters_n_hidden_shared = [20]
        self.holtwinters_n_hidden_cat_combo = [[2], [3], [5], [3]]
        self.holtwinters_n_hidden_num = []
        self.arima_n_hidden_shared = [40]
        self.arima_n_hidden_cat_combo = [[5], [5], [5]]
        self.arima_n_hidden_num = []
        self.sarima_n_hidden_shared = [40]
        self.sarima_n_hidden_cat_combo = [[5], [5], [5], [5], [5]]
        self.sarima_n_hidden_num = []
        self.theta_n_hidden_shared = [40]
        self.theta_n_hidden_cat_combo = [[5]]
        self.theta_n_hidden_num = []
        self.stlf_n_hidden_shared = [20]
        self.stlf_n_hidden_cat_combo = [[5], [5]]
        self.stlf_n_hidden_num = []
        self.cusum_n_hidden_shared = [20]
        self.cusum_n_hidden_cat_combo = [[3]]
        self.cusum_n_hidden_num = [5, 5, 5]
        self.statsig_n_hidden_shared = [20]
        self.statsig_n_hidden_cat_combo = []
        self.statsig_n_hidden_num = [5, 5]


def _log_error(msg: str) -> ValueError:
    LOGGER.error(msg)
    return ValueError(msg)


class MetaLearnHPT:
    """A class for meta-learner framework on hyper-parameters tuning.

    MetaLearnHPT is a framework for choosing hyper-parameter for forecasting models. It uses a multi-task neural network for recommendation, with time series features as inputs, and the hyper-parameters of a given model as outputs.
    For training, it uses time series features as inputs and the corresponding best hyper-parameters as labels. For prediction, it takes time series or time series features to predict the best hyper-parameters.
    MetaLearnHPT provides get_default_model, build_network, train, pred, pred_by_feature, save_model, load_model and plot.

    Attributes:
        data_x: Optional; A `pandas.DataFrame` object of time series features. data_x should not be None unless load_model is True. Default is None.
        data_y: Optional; A `pandas.DataFrame` object of the corresponding best hyper-parameters. data_y should not be None unless load_model is True. Default is None.
        categorical_idx: Optional; A list of strings of the names of the categorical hyper-parameters. Default is None.
        numerical_idx: Optional; A list of strings of the names of the numerical hyper-parameters. Default is None.
        default_model: Optional; A string of the name of the forecast model whose default settings will be used.
                       Can be 'arima', 'sarima', 'theta', 'holtwinters', 'stlf' or None. Default is None.
        scale: Optional; A boolean to specify whether or not to normalize time series features to zero mean and unit variance. Default is True.
        load_model: Optional; A boolean to specify whether or not to load a trained model. Default is False.

    Sample Usage:
        >>> mlhpt_hw = MetaLearnHPT(X, Y, default_model='holtwinters') # Use default Holt-Winter's model as an example.
        >>> mlhpt_hw.build_network()
        >>> mlhpt_hw.train()
        >>> mlhpt_hw.pred(ts=TSdata) # Recommend hyper-parameters for TSdata.
        >>> mlhpt_hw.save_model('my_model_binary.pkl') # Save trained model to a binary
        >>> mlhpt_hw2=MetaLearnHPT(load_model=True) # Load a trained model
        >>> mlhpt_hw2.load_model('my_model_binary.pkl')
        >>> mlhpt_hw = MetaLearnHPT(X, Y_holtwinters, ['trend','damped', 'seasonal'], ['seasonal_periods']) # Example for building customized MetaLearnHPT object.
        >>> mlhpt_hw.build_network(n_hidden_shared=[30], n_hidden_cat_combo=[[2], [3], [5]],n_hidden_num=[3])
        >>> mlhpt_hw.train(loss_scale=30, lr=0.001)
    """

    def __init__(
        self,
        data_x: Optional[pd.DataFrame] = None,
        data_y: Optional[pd.DataFrame] = None,
        categorical_idx: Optional[List[str]] = None,
        numerical_idx: Optional[List[str]] = None,
        default_model: Optional[str] = None,
        scale: bool = True,
        load_model: bool = False,
        # pyre-fixme[2]: Parameter must be annotated.
        **kwargs,
    ) -> None:
        if not load_model:
            if data_x is None:
                raise _log_error("data_x is necessary to initialize a new model!")
            if data_y is None:
                raise _log_error("data_y is necessary to initialize a new model!")

            data_x.fillna(0, inplace=True)
            # pyre-fixme[4]: Attribute must be annotated.
            self.dataX = np.asarray(data_x)
            # pyre-fixme[4]: Attribute must be annotated.
            self.dataY = data_y.copy()
            # pyre-fixme[4]: Attribute must be annotated.
            self.dim_input = self.dataX.shape[1]
            # pyre-fixme[4]: Attribute must be annotated.
            self.model = None

            # Record loss path for validation/trainin set and for both classification and regression.
            # pyre-fixme[4]: Attribute must be annotated.
            self._loss_path = collections.defaultdict(list)

            if isinstance(default_model, str):
                default_model = default_model.lower()
            # pyre-fixme[4]: Attribute must be annotated.
            self.__default_model = default_model

            default_model_params = DefaultModelParams()

            if default_model is not None:

                if (categorical_idx is not None) or (numerical_idx is not None):
                    msg = """
                         Default model cannot accept customized categorical_idx or customized numerical_idx! Please set
                         'categorical_idx=None' and 'numerical_idx=None' to initialize a default model,
                         or set 'default_model=None' to initialize a customized model!
                         """
                    raise _log_error(msg)

                if default_model in _MODELS:
                    categorical_idx_var = f"{default_model}_categorical_idx"
                    numerical_idx_var = f"{default_model}_numerical_idx"
                    categorical_idx = getattr(default_model_params, categorical_idx_var)
                    numerical_idx = getattr(default_model_params, numerical_idx_var)
                else:
                    msg = f"default_model={default_model} is not available! Please choose one from 'arima', 'sarima', 'holtwinters', 'stlf', 'theta', 'cusum', 'statsig'"
                    raise _log_error(msg)

            if (not numerical_idx) and (not categorical_idx):
                msg = "At least one of numerical_idx and categorical_idx should be a non-empty list."
                raise _log_error(msg)

            # pyre-fixme[4]: Attribute must be annotated.
            self.categorical_idx = categorical_idx
            # pyre-fixme[4]: Attribute must be annotated.
            self.numerical_idx = numerical_idx
            # pyre-fixme[4]: Attribute must be annotated.
            self._target_num = np.asarray(self.dataY[self.numerical_idx]) if self.numerical_idx else None
            # pyre-fixme[4]: Attribute must be annotated.
            self._dim_output_num = self._target_num.shape[1] if self.numerical_idx else 0
            self._get_target_cat()
            self._validate_data()
            # pyre-fixme[4]: Attribute must be annotated.
            self.scale = scale
            if self.scale:
                # pyre-fixme[4]: Attribute must be annotated.
                self.x_mean = self.dataX.mean(0)
                x_std = self.dataX.std(0)
                x_std[x_std == 0.0] = 1.0
                # pyre-fixme[4]: Attribute must be annotated.
                self.x_std = x_std
                self.dataX = (self.dataX - self.x_mean) / self.x_std

        self.n_hidden_shared = [0]
        self.n_hidden_cat_combo = [0]
        self.n_hidden_num = [0]

    def _get_target_cat(self) -> None:
        # List of number of classes (dim of output) of each categorical variable
        if self.categorical_idx is None:
            self.target_cat = None
            self.dim_output_cat = []
            return
        n_cat = []
        # Dict for encoder, categories --> int
        self.cat_code_dict = {}
        for col in self.categorical_idx:
            n_cat.append(self.dataY[col].nunique())
            self.dataY[col] = self.dataY[col].astype("category")
            self.cat_code_dict[col] = dict(enumerate(self.dataY[col].cat.categories))
            self.dataY[col] = self.dataY[col].cat.codes.values

        self.target_cat = np.asarray(self.dataY[self.categorical_idx]) if self.categorical_idx else None
        self.dim_output_cat = n_cat

    def get_default_model(self) -> Optional[str]:
        """Get the name of default_model. It the instance is a customized model, return None.

        Returns:
            A string reprsenting the default model or None.
        """

        return self.__default_model

    def _validate_data(self) -> None:
        """Validate input data."""

        n_cat = len(self.categorical_idx) if self.categorical_idx is not None else 0
        n_num = len(self.numerical_idx) if self.numerical_idx is not None else 0
        dim = self.dataY.shape[1]
        if n_cat + n_num != dim:
            msg = f"Dimensions of data_y (dim={dim}) and the input variables (dim={n_cat}+{n_num}) do not agree!"
            raise _log_error(msg)

        for i, var in enumerate(self.categorical_idx):
            if self.dim_output_cat[i] == 1:
                msg = f"Column {var} only has one class, not able to train a model!"
                raise _log_error(msg)

        if self.dataX.shape[0] <= 30:
            msg = "Dataset is too small to train a model!"
            raise _log_error(msg)

    @staticmethod
    def _get_hidden_and_output_cat_combo(
        n_hidden_cat_combo: List[List[int]], out_dim_cat: List[int]
    ) -> List[List[int]]:
        # The length of n_hidden_cat_combo and out_dim_cat should be same
        # If there is no categorical variable, out_dim_cat = []
        if not out_dim_cat:
            return []
        res = []
        for i in range(len(n_hidden_cat_combo)):
            res.append(n_hidden_cat_combo[i] + [out_dim_cat[i]])
        return res

    @staticmethod
    def _get_hidden_and_output_num(n_hidden_num: List[int], out_dim_num: int) -> List[int]:
        # If there is no numerical variable, out_dim_num = []
        if not out_dim_num:
            return []
        return n_hidden_num + [out_dim_num]

    def build_network(
        self,
        n_hidden_shared: Optional[List[int]] = None,
        n_hidden_cat_combo: Optional[List[List[int]]] = None,
        n_hidden_num: Optional[List[int]] = None,
    ) -> None:
        """Build a multi-task neural network.

        This function builds a multi-task neural network according to given neural network structure (i.e., n_hidden_shared, n_hidden_cat_combo, n_hidden_num).
        If the MetaLearnHPT object is initiated as a default_model (i.e., default_model is not None), then a default neural network structure will be built and cannot accept customized.

        Args:
            n_hidden_shared: Optional; A list of numbers of hidden neurons in each shared hidden layer.
                             For example, n_hidden_shared = [first_shared_hid_dim, second_shared_hid_dim, ....]. Default is None.

            n_hidden_cat_combo: Optional; A list of lists of task-specific hidden layers’ sizes of each categorical response variables.
                                For example, if we have 3 categorical y, then
                                n_hidden_cat_combo = [[first_spec_hid_dim_cat1, second_spec_hid_dim_cat1, ...],
                                [first_spec_hid_dim_cat2, second_spec_hid_dim_cat2, ...],
                                [first_spec_hid_dim_cat3, second_spec_hid_dim_cat3, ...]].
                                Length of n_hidden_cat_combo must match the number of categorical y. Default is None.

            n_hidden_num: Optional; A list of task-specific hidden layers’ sizes of numerical response variables.
                          For example, n_hidden_num = [first_spec_hid_dim, second_spec_hid_dim, ...]. Default is None.

        Returns:
            None.
        """

        network_structure = (n_hidden_shared is None) and (n_hidden_cat_combo is None) and (n_hidden_num is None)

        default_model = self.__default_model
        default_model_networks = DefaultModelNetworks()
        if default_model is not None:
            if not network_structure:
                msg = f"A default model structure ({default_model}) is initiated and cannot accept the customized network structure!"
                raise _log_error(msg)

            if default_model in _MODELS:
                n_hidden_shared_var = f"{default_model}_n_hidden_shared"
                n_hidden_cat_combo_var = f"{default_model}_n_hidden_cat_combo"
                n_hidden_num_var = f"{default_model}_n_hidden_num"
                n_hidden_shared = getattr(default_model_networks, n_hidden_shared_var)
                n_hidden_cat_combo = getattr(default_model_networks, n_hidden_cat_combo_var)
                n_hidden_num = getattr(default_model_networks, n_hidden_num_var)
            else:
                msg = f"Default neural network for model {default_model} is not implemented!"
                raise _log_error(msg)
            msg = f"Default neural network for model {default_model} is built."
            LOGGER.info(msg)
        elif n_hidden_shared is None:
            msg = "n_hidden_shared is missing!"
            raise _log_error(msg)
        elif n_hidden_cat_combo is None:
            msg = "n_hidden_cat_combo is missing!"
            raise _log_error(msg)
        elif n_hidden_num is None:
            msg = "n_hidden_num is missing!"
            raise _log_error(msg)

        if len(n_hidden_cat_combo) != len(self.dim_output_cat):
            msg = "Unmatched dimension!"
            raise _log_error(msg)
        # Add input dim before n_hidden_shared.
        # Add output dim at the end of n_hidden_cat_combo.
        # Add output dim at the end of n_hidden_num.
        self.n_hidden_shared = n_hidden_shared
        self.n_hidden_cat_combo = n_hidden_cat_combo
        self.n_hidden_num = n_hidden_num

        self.model = MultitaskNet(
            input_and_n_hidden_shared=[self.dim_input] + n_hidden_shared,
            n_hidden_and_output_cat_combo=self._get_hidden_and_output_cat_combo(
                n_hidden_cat_combo, self.dim_output_cat
            ),
            n_hidden_and_output_num=self._get_hidden_and_output_num(n_hidden_num, self._dim_output_num),
        )
        print("Multi-task neural network structure:")
        print(self.model)

    def _prepare_data(
        self, val_size: float
    ) -> Tuple[
        torch.FloatTensor,
        Optional[torch.LongTensor],
        Optional[torch.FloatTensor],
        torch.FloatTensor,
        Optional[torch.LongTensor],
        Optional[torch.FloatTensor],
    ]:
        # split to train and validation sets
        train_idx, val_idx = train_test_split(np.arange(len(self.dataX)), test_size=val_size)

        # change training set to tensors
        x_fs = torch.from_numpy(self.dataX[train_idx, :]).float()
        y_cat = torch.from_numpy(self.target_cat[train_idx, :]).long() if self.categorical_idx else None
        y_num = (
            torch.from_numpy(self._target_num[train_idx, :].astype("float")).float() if self.numerical_idx else None
        )

        # change validation set to tensors
        x_fs_val = torch.from_numpy(self.dataX[val_idx, :]).float()
        y_cat_val = torch.from_numpy(self.target_cat[val_idx, :]).long() if self.categorical_idx else None
        y_num_val = (
            torch.from_numpy(self._target_num[val_idx, :].astype("float")).float() if self.numerical_idx else None
        )
        # pyre-fixme[7]: Expected `Tuple[FloatTensor, Optional[LongTensor],
        #  Optional[FloatTensor], FloatTensor, Optional[LongTensor],
        #  Optional[FloatTensor]]` but got `Tuple[Tensor, Optional[Tensor],
        #  Optional[Tensor], Tensor, Optional[Tensor], Optional[Tensor]]`.
        return x_fs, y_cat, y_num, x_fs_val, y_cat_val, y_num_val

    def _loss_function(
        self,
        o1: Optional[torch.Tensor],
        o2: Optional[torch.Tensor],
        y_cat: Optional[torch.Tensor],
        y_num: Optional[torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Loss function for multi-task neural network."""

        loss_func_num = nn.MSELoss()
        loss_func_cat = nn.CrossEntropyLoss()
        loss_cat_train = torch.tensor([0.0])
        # Loss of classification.
        if o1 is not None:
            assert y_cat is not None
            batch_y1 = y_cat
            for col in range(batch_y1.shape[1]):
                loss_cat_train += loss_func_cat(o1[col], batch_y1[:, col])

        # Loss of regression.
        loss_num_train = torch.tensor([0.0])
        if o2 is not None:
            batch_y2 = y_num
            loss_num_train += loss_func_num(o2, batch_y2)

        return loss_cat_train, loss_num_train

    def train(
        self,
        loss_scale: float = 1.0,
        lr: float = 0.001,
        n_epochs: int = 1000,
        batch_size: int = 128,
        method: str = "SGD",
        val_size: float = 0.1,
        momentum: float = 0.9,
        n_epochs_stop: Union[int, float] = 20,
    ) -> None:
        """Train the pre-built multi-task neural network.

        Args:
            loss_scale: Optional; A float to specify the hyper-parameter to scale regression loss and classification loss, which controls the trade-off between the accuracy of regression task and classification task.
                        A larger loss_scale value gives a more accurate prediction for classification part, and a lower value gives a more accurate prediction for regression part. Default is 1.0.
            lr: Optional; A float for learning rate. Default is 0.001.
            n_epochs: Optional; An integer for the number of epochs. Default is 1000.
            batch_size: Optional; An integer for the batch size. Default is 128.
            method: Optional; A string for the name of optimizer. Can be 'SGD' or 'Adam'. Default is 'SGD'.
            val_size: Optional; A float for the proportion of validation set of. It should be within (0, 1). Default is 0.1.
            momentum: Optional; A fload for the momentum for SGD. Default value is 0.9.
            n_epochs_stop: Optional; An integer or a float for early stopping condition. If the number of epochs is larger than n_epochs_stop and there is no improvement on validation set, we stop training.
                           One can turn off the early stopping feature by setting n_epochs_stop = np.inf. Default is 20.

        Returns:
            None.
        """

        if self.model is None:
            raise _log_error("Haven't built a model. Please build a model first!")

        if method == "SGD":
            optimizer = torch.optim.SGD(self.model.parameters(), lr=lr, momentum=momentum)
        elif method == "Adam":
            optimizer = torch.optim.Adam(self.model.parameters(), lr=lr)
        else:
            raise _log_error("Only support SGD and Adam optimaizer. Please use 'SGD' or 'Adam'.")

        if val_size <= 0 or val_size >= 1:
            raise _log_error("Illegal validation size.")

        # get training tensors
        x_fs, y_cat, y_num, x_fs_val, y_cat_val, y_num_val = self._prepare_data(val_size)

        # validate batch size
        if batch_size >= x_fs.size()[0]:
            raise _log_error(f"batch_size {batch_size} is larger than training data size {x_fs.size()[0]}!")

        # variables for early stopping
        min_val_loss = np.inf
        epochs_no_improve = 0.0

        for epoch in range(n_epochs):
            total_loss = 0
            permutation = torch.randperm(x_fs.size()[0])
            for i in range(0, x_fs.size()[0], batch_size):
                indices = permutation[i : i + batch_size]
                batch_x = x_fs[indices]

                # two outputs, o1: classification, o2: regression
                o1, o2 = self.model.forward(batch_x)
                tmp_y_cat = y_cat[indices] if y_cat is not None else None
                tmp_y_num = y_num[indices] if y_num is not None else None
                loss_cat_train, loss_num_train = self._loss_function(o1, o2, tmp_y_cat, tmp_y_num)
                cur_loss = loss_cat_train + loss_num_train / loss_scale
                optimizer.zero_grad()
                cur_loss.backward()
                optimizer.step()

                total_loss += cur_loss

            # Record loss of training set at the last iteration of each epoch.
            self._loss_path["LOSS_train_cat"].append(loss_cat_train.item())
            self._loss_path["LOSS_train_num"].append(loss_num_train.item())

            # Record loss of validatiaon set for each epoch.
            o1_val, o2_val = self.model.forward(x_fs_val)
            loss_cat_val, loss_num_val = self._loss_function(o1_val, o2_val, y_cat_val, y_num_val)
            self._loss_path["LOSS_val_cat"].append(loss_cat_val.item())
            self._loss_path["LOSS_val_num"].append(loss_num_val.item())
            loss_sum_val = loss_cat_val + loss_num_val / loss_scale

            # early stopping variables update
            if loss_sum_val < min_val_loss:
                min_val_loss = loss_sum_val
                epochs_no_improve = 0.0
            else:
                epochs_no_improve += 1

            # check early stopping condition
            if epoch > 20 and epochs_no_improve >= n_epochs_stop:
                LOGGER.info(f"Early stopping! Stop at epoch {epoch + 1}.")
                break

    def pred(self, source_ts: TimeSeriesData, ts_scale: bool = True) -> pd.DataFrame:
        """Predict hyper-parameters for a new time series data.

        Args:
            source_ts: :class:`kats.consts.TimeSeriesData` object representing the time series for which to generate hyper-parameters
            ts_scale: A boolean to specify whether or not to rescale time series data (i.e., divide its value by its maximum value) before calculating its features. Default is True.

        Returns:
            A `pandas.DataFrame` object storing the recommended hyper-parameters.
        """

        ts = TimeSeriesData(pd.DataFrame(source_ts.to_dataframe().copy()))

        if self.model is None:
            raise _log_error("Haven't trained a model. Please train a model or load a model before predicting.")

        if ts_scale:
            # scale time series to make ts features more stable
            ts.value /= ts.value.max()
            msg = "Successful scaled! Each value of TS has been divided by the max value of TS."
            LOGGER.info(msg)

        self.model.eval()
        new_feature = TsFeatures().transform(ts)
        # pyre-fixme[16]: `List` has no attribute `values`.
        new_feature_vector = np.asarray(list(new_feature.values()))

        if np.any(np.isnan(new_feature_vector)):
            LOGGER.warning(
                "Time series features contain NaNs!" f"Time series features are {new_feature}. " "Fill in NaNs with 0."
            )

        pred_res = self.pred_by_feature([new_feature_vector])[0]

        # To have a consistent type with orginal HPT methods' output.
        res = [["0_0", "unknown", 0, 0.0, 0, pred_res]]
        res = pd.DataFrame(res, copy=False)
        res.columns = [
            "arm_name",
            "metric_name",
            "mean",
            "sem",
            "trial_index",
            "parameters",
        ]
        return res

    def pred_by_feature(self, source_x: Union[np.ndarray, List[np.ndarray], pd.DataFrame]) -> List[Dict[str, Any]]:
        """Predict hyper-parameters for time series features.

        Args:
            source_x: Time series features.

        Returns:
            A list of dictionaries storing the recommended hyper-parameters.
        """

        if self.model is None:
            raise _log_error("Haven't trained a model. Please train a model or load a model before predicting.")
        if isinstance(source_x, List):
            x = np.row_stack(source_x)
        elif isinstance(source_x, pd.DataFrame):
            x = source_x.values.copy()
        elif isinstance(source_x, np.ndarray):
            x = source_x.copy()
        else:
            raise _log_error(f"In valid source_x type: {type(source_x)}.")
        x[np.isnan(x)] = 0.0
        if self.scale:
            x = (x - self.x_mean) / self.x_std

        n = len(x)
        x = torch.from_numpy(x).float()

        self.model.eval()
        cats, nums = self.model(x)
        cats = [torch.argmax(t, dim=1).detach().numpy() for t in cats] if cats is not None else []
        nums = nums.detach().numpy() if nums is not None else []

        ans = [{} for _ in range(n)]
        for j, c in enumerate(self.categorical_idx):
            vals = cats[j]
            for i in range(n):
                ans[i][c] = self.cat_code_dict[c][vals[i]]
        for j, c in enumerate(self.numerical_idx):
            vals = nums[:, j]
            for i in range(n):
                ans[i][c] = vals[i]
        return ans

    def save_model(self, file_path: str) -> None:
        """Save trained model to a binary.

        Args:
            file_path: A string representing the path to save a trained model, which should contain either '.p' or '.pkl' file extension.

        Returns:
            None.
        """

        if self.model is None:
            raise _log_error("Haven't trained a model.")
        else:
            joblib.dump(self.__dict__, file_path)
            LOGGER.info("Successfully saved the trained model!")

    def load_model(self, file_path: str) -> None:
        """Load a pre-trained model from a binary.

        Args:
            file_path: A string representing the path to load a pre-trained model.

        Returns:
            None.
        """

        try:
            self.__dict__ = joblib.load(file_path)
        except Exception as e:
            raise _log_error(f"Fail to load model from {file_path}, and error message is: {e}")

    def plot(
        self,
        figsize: Optional[Tuple[int, int]] = None,
        xlabel: str = "Epoch",
        train_linestyle: str = ".-",
        validation_linestyle: str = "o-",
        classification_title: str = "Loss path of classification tasks",
        regression_title: str = "Loss path of regression task",
        classification_ylabel: str = "Cross-entropy",
        regression_ylabel: str = "MSE",
        use_legend: bool = True,
        legend: Optional[List[str]] = None,
    ) -> Sequence[plt.Axes]:
        """Plot loss paths of classification/regression on both training and validation.

        Args:
            figsize: optional figure size if creating. If None, uses (15, 7).
            xlabel: x-axis label.
            train_linestyle: line style to use for training.
            validation_linestyle: line style to use for validation.
            classification_title: title for classification plot.
            regression_title: title for regression plot.
            classification_ylabel: y-axis label for classification plot.
            regression_ylabel: y-axis label for regression plot.
            use_legend: if True, plots legend on both plots.
            legend: Legend labels to use. if None, uses ["training set",
                "validation set"].

        Returns:
            The matplotlib Axes.
        """

        if not self._loss_path["LOSS_train_cat"] and not self._loss_path["LOSS_train_num"]:
            raise _log_error("Using a loaded model or no trained model!")

        if figsize is None:
            figsize = (15, 7)
        _, axs = plt.subplots(1, 2, figsize=figsize, sharex=True)
        if legend is None:
            legend = ["training set", "validation set"]

        ax = axs[0]
        ax.plot(self._loss_path["LOSS_train_cat"], train_linestyle)
        ax.plot(self._loss_path["LOSS_val_cat"], validation_linestyle)
        if use_legend:
            ax.legend(legend)
        ax.set_title(classification_title)
        ax.set_ylabel(classification_ylabel)

        ax = axs[1]
        ax.plot(self._loss_path["LOSS_train_num"], train_linestyle)
        ax.plot(self._loss_path["LOSS_val_num"], validation_linestyle)
        if use_legend:
            ax.legend(legend)
        ax.set_title(regression_title)
        ax.set_ylabel(regression_ylabel)
        ax.set_xlabel(xlabel)
        return axs


class MultitaskNet(nn.Module):
    """A class for multi-task neural network.

    Build a multi-task neural network used by MetaLearnHPT. It can also be used to single-task learning. Currently only support Relu activation function.

    Attributes:
        input_and_n_hidden_shared: A list of integers contains the dimension of input and numbers of hidden neurons in each shared hidden layer.
                                   The first value in this list is dimension of input, which is dimension of feature vector.
                                   For example, input_and_n_hidden_shared = [input_dim, first_shared_hid_dim, second_shared_hid_dim, ....].
        n_hidden_and_output_cat_combo: A list of lists of task-specific hidden layers’ sizes of each categorical response variables and their dimension of output.
                                       For example, if we have 3 categorical y with 3, 2, 4 classes respectively, then
                                       n_hidden_and_output_cat_combo = [[first_spec_hid_dim_cat1, second_spec_hid_dim_cat1, ..., 3],
                                       [first_spec_hid_dim_cat2, second_spec_hid_dim_cat2, ..., 2],
                                       [first_spec_hid_dim_cat3, second_spec_hid_dim_cat3, ..., 4]].
        n_hidden_and_output_num: A list of integers contains task-specific hidden layers’ sizes of numerical response variables and the dimension of output,
                                 which is the number of numerical response variables in data_y.
                                 For example, if we have three numerical response variables in y, then
                                 n_hidden_and_output_num = [first_spec_hid_dim, second_spec_hid_dim, ..., 3]
    """

    def __init__(
        self,
        input_and_n_hidden_shared: List[int],
        n_hidden_and_output_cat_combo: List[List[int]],
        n_hidden_and_output_num: List[int],
    ) -> None:
        super(MultitaskNet, self).__init__()
        self.shared_layer = nn.ModuleList()
        for i in range(len(input_and_n_hidden_shared) - 1):
            self.shared_layer.append(nn.Linear(input_and_n_hidden_shared[i], input_and_n_hidden_shared[i + 1]))

        self.cat_layer_combo = nn.ModuleList()
        for n_hidden_and_output_cat in n_hidden_and_output_cat_combo:
            cat_layer = nn.ModuleList()
            # input for task-specific layer is the dim of last shared hidden layer
            curr_input = input_and_n_hidden_shared[-1]
            for i in range(len(n_hidden_and_output_cat)):
                cat_layer.append(nn.Linear(curr_input, n_hidden_and_output_cat[i]))
                curr_input = n_hidden_and_output_cat[i]
            self.cat_layer_combo.append(cat_layer)

        self.num_layer = nn.ModuleList()
        curr_input = input_and_n_hidden_shared[-1]
        for i in range(len(n_hidden_and_output_num)):
            self.num_layer.append(nn.Linear(curr_input, n_hidden_and_output_num[i]))
            curr_input = n_hidden_and_output_num[i]

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def forward(self, x):
        """Forward function in neural networks."""

        # Shared layers.
        for layer in self.shared_layer:
            x = layer(x)
            x = nn.functional.relu(x)

        # Categorical part.
        if not self.cat_layer_combo:
            y_pred_cat_combo = None
        else:
            y_pred_cat_combo = []
            for cat_layer in self.cat_layer_combo:
                y_pred_cat = cat_layer[0](x)
                for i in range(1, len(cat_layer)):
                    # the last layer has no activation function
                    y_pred_cat = nn.functional.relu(y_pred_cat)
                    y_pred_cat = cat_layer[i](y_pred_cat)
                y_pred_cat_combo.append(y_pred_cat)

        # Numerical part.
        if not self.num_layer:
            y_pred_num = None
        else:
            y_pred_num = self.num_layer[0](x)
            for i in range(1, len(self.num_layer)):
                # the last layer has no activation function
                y_pred_num = nn.functional.relu(y_pred_num)
                y_pred_num = self.num_layer[i](y_pred_num)

        return y_pred_cat_combo, y_pred_num
